"""create_local_train_profile_table

Revision ID: c1ee06d84c12
Revises: c22620b97e39
Create Date: 2021-06-07 17:43:49.859860

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'c1ee06d84c12'
down_revision = 'c22620b97e39'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('local_train_profile',
    sa.Column('id', sa.String(length=32), nullable=False),
    sa.Column('train_task_id', sa.String(length=32), nullable=True),
    sa.Column('algorithm_id', sa.String(length=32), nullable=False),
    sa.Column('algorithm_dir', sa.String(length=4096), nullable=True),
    sa.Column('algorithm_archive', sa.String(length=4096), nullable=True),
    sa.Column('dataset_id', sa.String(length=32), nullable=False),
    sa.Column('dataset_dir', sa.String(length=4096), nullable=True),
    sa.Column('dataset_archive', sa.String(length=4096), nullable=True),
    sa.Column('entrypoint', sa.String(length=4096), nullable=True),
    sa.Column('train_params', sa.String(length=4096), nullable=True),
    sa.Column('created_at', sa.DateTime(), nullable=False),
    sa.ForeignKeyConstraint(['train_task_id'], ['local_train_task.id'], ondelete='SET NULL'),
    sa.PrimaryKeyConstraint('id')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('local_train_profile')
    # ### end Alembic commands ###



import os
from pathlib import Path
from typing import Dict, List

import click
from ruamel.yaml import YAML

from anylearn.applications.quickstart import quick_train
from anylearn.cli.anylearn_cli_config import AnylearnCliConfig
from anylearn.cli.utils import (
    check_config,
    check_connection,
    cmd_confirm_or_abort,
    cmd_error,
    cmd_info,
    cmd_success,
    cmd_warning,
    get_cmd_command,
)
from anylearn.interfaces import TrainTask


@click.group("run")
def commands():
    """
    Run training, evaluation or serving on remote Anylearn.
    """
    pass


@commands.command()
@click.argument('algorithm_name')
@click.option(
    '-d', '--data',
    multiple=True,
    help="""
    Dataset(s) KV pair in `<key>=<dataset-name>` format.
    """
)
@click.option(
    '--data-yaml',
    help="""
    Absolute path of datasets KV definition file in YAML format.
    If specified, datasets through `-d|--data` option will be ignored.
    """
)
@click.option(
    '-p', '--param',
    multiple=True,
    help="""
    Hyperparam(s) KV pair in `<key>=<value>` format.
    """
)
@click.option(
    '--param-yaml',
    help="""
    Absolute path of hyperparams KV definition file in YAML format.
    If specified, hyperparams through `-p|--param` option will be ignored.
    """
)
@click.option(
    '-g', '--quota-group',
    default='default',
    help="""
    Specify a certain QuotaGroup to request computing resource within.
    """
)
@click.option(
    '-r', '--resource',
    multiple=True,
    help="""
    Computing resource(s) KV pair in `<key>=<value>` format.
    """
)
@click.option(
    '--resource-yaml',
    help="""
    Absolute path of resource request K-KV definition file in YAML format.
    If specified, resource request through `-g|--quota-group` and `-r|--resource`
    options will be ignored.
    """
)
@click.option(
    '--save',
    is_flag=True,
    default=False,
    help="""
    Whether save the training as reusable configuration file.
    """
)
@check_config()
@check_connection()
@get_cmd_command()
def training(
    algorithm_name: str,
    data: tuple,
    param: tuple,
    resource: tuple,
    data_yaml: str=None,
    param_yaml: str=None,
    resource_yaml: str=None,
    quota_group: str='default',
    save: bool=False,
):
    """
    Run training task.
    Note that algorithm to train is selected by name.
    Therefore, please maintain unique algorithm names in current project's scope.
    """
    config = AnylearnCliConfig.load()
    _check_project(config)
    _check_algorithm_name(config, algorithm_name)
    data_config = _get_data_config(list(data), data_yaml)
    param_config = _get_param_config(list(param), param_yaml)
    resource_config = _get_resource_config(quota_group,
                                           list(resource),
                                           resource_yaml)
    kwargs = {
        'algorithm_name': algorithm_name,
        'data_config': data_config,
        'param_config': param_config,
        'resource_config': resource_config,
    }
    if save:
        path = _dump_training_config(**kwargs)
        cmd_info(msg=f"Training config saved in {path}")
    task = _do_train(config=config, **kwargs)
    cmd_success(msg=f"CREATED {task.id}")


def _check_project(config: AnylearnCliConfig):
    if not config.project or not config.project.id:
        cmd_error(msg=(
            "Local project is broken or has not been published to Anylearn remote.\n"
            "Call `anyctl reset` then `anyctl init` if project configuration is corrupted, "
            f"or call `anyctl push project` to publish local project."
        ))
        raise click.Abort


def _check_algorithm_name(config: AnylearnCliConfig, algorithm_name: str):
    try:
        algo = config.algorithms[algorithm_name]
        _ = config.path['algorithm'][algorithm_name]
    except KeyError:
        cmd_error(msg=(
            f"Algorithm {algorithm_name} does not exist in current project."
        ))
        raise click.Abort
    if not algo.id:
        cmd_error(msg=(
            f"Algorithm {algorithm_name} has not been pushed to Anylearn remote.\n"
            f"Call `anyctl push algorithm {algorithm_name}` first."
        ))
        raise click.Abort


def _check_dataset_name(config: AnylearnCliConfig, dataset_name: str):
    try:
        dset = config.datasets[dataset_name]
        _ = config.path['dataset'][dataset_name]
    except KeyError:
        cmd_error(msg=(
            f"Dataset named {dataset_name} does not exist in current project."
        ))
        raise click.Abort
    if not dset.id:
        cmd_error(msg=(
            f"Dataset {dataset_name} has not been pushed to Anylearn remote.\n"
            f"Call `anyctl push dataset {dataset_name}` first."
        ))
        raise click.Abort


def _get_data_config(data_kv: List[str], data_yaml: str) -> Dict[str, str]:
    if data_yaml:
        yaml = YAML()
        yaml_path = Path(data_yaml)
        if not yaml_path.exists():
            cmd_error(msg=f"No such file: {data_yaml}.")
            raise click.Abort
        return yaml.load(yaml_path)
    elif data_kv and isinstance(data_kv, list):
        return dict(kv.split('=') for kv in data_kv)
    else:
        return {}


def _get_param_config(param_kv: List[str], param_yaml: str) -> Dict[str, str]:
    if param_yaml:
        yaml = YAML()
        yaml_path = Path(param_yaml)
        if not yaml_path.exists():
            cmd_error(msg=f"No such file: {param_yaml}.")
            raise click.Abort
        return yaml.load(yaml_path)
    elif param_kv and isinstance(param_kv, list):
        return dict(kv.split('=') for kv in param_kv)
    else:
        return {}


def _get_resource_config(quota_group: str,
                         resource_kv: List[str],
                         resource_yaml: str) -> Dict[str, str]:
    if resource_yaml:
        yaml = YAML()
        yaml_path = Path(resource_yaml)
        if not yaml_path.exists():
            cmd_error(msg=f"No such file: {resource_yaml}.")
            raise click.Abort
        return yaml.load(yaml_path)
    elif quota_group:
        req = {}
        for kv in resource_kv:
            rk, rv = kv.split('=')
            try:
                req[rk] = int(rv)
            except ValueError:
                cmd_error(msg=(f"Resource quantity is not integer: {kv}."))
                raise click.Abort
        return [{quota_group: req}]
    else:
        return {}


def _do_train(config: AnylearnCliConfig,
              algorithm_name: str,
              data_config: Dict[str, str],
              param_config: Dict[str, str],
              resource_config: Dict[str, str]) -> TrainTask:
    algo = config.algorithms[algorithm_name]
    algo_dir = config.path['algorithm'][algorithm_name]
    datasets = {}
    for k, v in data_config.items():
        _check_dataset_name(config, v)
        dset = config.datasets[v]
        dset_dir = config.path['dataset'][v]
        datasets[k] = {
            'val': f"${dset.id}",
            'dir': dset_dir,
        }
    # Sum up and call `quick_train`
    kwargs = {
        'project_id': config.project.id,
        'algorithm_id': algo.id,
        'algorithm_dir': algo_dir,
        'hyperparams': param_config,
        'resource_request': resource_config,
    }
    if datasets:
        # TODO: support multiple datasets in quick_train
        kwargs['dataset_hyperparam_name'] = list(datasets.keys())[0]
        kwargs['dataset_id'] = list(datasets.values())[0]['val'].replace('$', '')
    task, _, _, _ = quick_train(**kwargs)
    return task


def _dump_training_config(algorithm_name: str,
                          data_config: Dict[str, str],
                          param_config: Dict[str, str],
                          resource_config: Dict[str, str]) -> Path:
    cwd = Path(os.getcwd())
    i = 0
    while (cwd / f"AnylearnTraining.{i}.yaml").exists():
        i += 1
    training_config_path = cwd / f"AnylearnTraining.{i}.yaml"
    training_config_path.touch()
    yaml = YAML()
    yaml.dump(
        {
            'algorithm_name': algorithm_name,
            'data_config': data_config,
            'param_config': param_config,
            'resource_config': resource_config,
        },
        training_config_path,
    )
    return training_config_path


def _load_training_config(training_config_path: str) -> dict:
    training_config_path = Path(training_config_path)
    if not training_config_path.exists():
        cmd_error(msg=f"No such file {training_config_path}.")
        raise click.Abort
    yaml = YAML()
    training_config = dict({
        'algorithm_name': None,
        'data_config': {},
        'param_config': {},
        'resource_config': {},
    }, **yaml.load(training_config_path))
    if not training_config['algorithm_name']:
        cmd_error(msg="Field 'algorithm_name' is required.")
        raise click.Abort
    return training_config


@commands.command()
@get_cmd_command()
def evaluation():
    """
    Coming soon.
    """
    pass


@commands.command()
@get_cmd_command()
def serving():
    """
    Coming soon.
    """
    pass



import responses

from anylearn.interfaces.mirror import Mirror
from anylearn.utils.errors import AnyLearnException, AnyLearnMissingParamException
from tests.base_test_case import BaseTestCase

class TestMirror(BaseTestCase):
    @responses.activate
    def test_list_mirror_ok(self):
        responses.add(responses.GET, url=self._url("mirror/list"),
                      json=[
                          {
                              'id': "MIRR001",
                              'name': "Test001",
                              'requirements': "test==0.0.1",
                          },
                          {
                              'id': "MIRR002",
                              'name': "Test002",
                              'requirements': "test==0.0.2",
                          },
                      ],
                      status=200)
        mirrors = Mirror.get_list()
        self.assertIsInstance(mirrors, list)
        self.assertEqual(len(mirrors), 2)
        self.assertIsInstance(mirrors[0], Mirror)
        self.assertIsInstance(mirrors[1], Mirror)
        self.assertEqual(mirrors[0].id, "MIRR001")
        self.assertEqual(mirrors[1].id, "MIRR002")
        self.assertEqual(mirrors[0].name, "Test001")
        self.assertEqual(mirrors[1].name, "Test002")

    @responses.activate
    def test_list_mirror_ko_unknown_response(self):
        responses.add(responses.GET, url=self._url("mirror/list"),
                      json={'msg': "Unknown response"},
                      status=200)
        with self.assertRaises(AnyLearnException) as ctx:
            Mirror.get_list()
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnException)
        self.assertEqual(e.msg, "请求未能得到有效响应")



import click
from requests import HTTPError
from typing import Optional, Union

from anylearn.applications.algorithm_manager import sync_algorithm
from anylearn.applications.dataset_manager import sync_dataset
from anylearn.cli.anylearn_cli_config import AnylearnCliConfig
from anylearn.cli.utils import (
    check_config,
    check_connection,
    cmd_confirm_or_abort,
    cmd_error,
    cmd_info,
    cmd_success,
    cmd_warning,
    get_cmd_command,
)
from anylearn.interfaces import Algorithm, Dataset, Project
from anylearn.interfaces.resource import (
    AsyncResourceUploader,
    SyncResourceUploader,
)
from anylearn.utils.errors import AnylearnRequiredLocalCommitException


_option_force = click.option(
    '-f', '--force',
    is_flag=True,
    default=False,
    help="Skip prompt and force actions."
)


_option_async_upload = click.option(
    '--async-upload',
    is_flag=True,
    default=False,
    help="Upload in asynchronous mode."
)


@click.group("push")
def commands():
    """
    Push local project or algorithm(s) or dataset(s) to remote Anylearn.
    """
    pass


@commands.command()
@_option_force
@_option_async_upload
@check_config()
@check_connection()
@get_cmd_command()
def all(force: bool=False, async_upload: bool=False):
    """
    Create/update local project, all algorithms and all datasets
    to remote Anylearn.
    """
    config = AnylearnCliConfig.load()
    # Project
    cmd_info(msg=f"Pushing project {config.project.name}...")
    config.project = _push_project(project=config.project, force=force)
    cmd_success(msg="PUSHED")
    AnylearnCliConfig.update(config)

    # All algorithms
    for name in config.algorithms.keys():
        try:
            cmd_info(msg=f"Pushing algorithm {name}...")
            algorithm = _push_1_algorithm(
                algorithm=config.algorithms[name],
                dir=config.path['algorithm'][name],
                image=config.images[name],
                force=force,
            )
            config.algorithms[name] = algorithm
            AnylearnCliConfig.update(config)
            cmd_success(msg="PUSHED")
        except KeyError:
            cmd_warning(msg=(
                f"Algorithm named {name} or its path config does not exist, "
                "ignored."
            ))
            continue
    # All datasets
    for name in config.datasets.keys():
        try:
            cmd_info(msg=f"Pushing dataset {name}...")
            dataset = _push_1_dataset(
                dataset=config.datasets[name],
                dir=config.path['dataset'][name],
                force=force,
            )
            config.datasets[name] = dataset
            AnylearnCliConfig.update(config)
            cmd_success(msg="PUSHED")
        except KeyError:
            cmd_warning(msg=(
                f"Dataset named {name} or its path config does not exist, "
                "ignored."
            ))
            continue
    cmd_success(msg="DONE")


@commands.command()
@_option_force
@check_config()
@check_connection()
@get_cmd_command()
def project(force: bool=False):
    """
    Create/update local project to remote Anylearn.
    """
    config = AnylearnCliConfig.load()
    config.project = _push_project(project=config.project, force=force)
    AnylearnCliConfig.update(config)
    cmd_success(msg="PUSHED")


def _push_project(project: Project, force: bool=False) -> Project:
    if project.id:
        try:
            remote_project = Project(project.id, load_detail=True)
            if remote_project == project:
                cmd_info(msg="Already up-to-date.")
                return project
            cmd_warning(msg=(
                "Remote project "
                "("
                f"id={remote_project.id}, "
                f"name={remote_project.name}"
                ") will be overridden."
            ))
            if not force:
                cmd_confirm_or_abort()
        except HTTPError:
            cmd_warning(msg=(
                "Remote project "
                "("
                f"id={project.id}, "
                f"name={project.name}"
                ") is unaccessible. "
                "A new project will be created."
            ))
            if not force:
                cmd_confirm_or_abort()
            project.id = None
    project.save()
    project.get_detail()
    return project


@commands.command()
@click.argument('name')
@_option_force
@_option_async_upload
@check_config()
@check_connection()
@get_cmd_command()
def algorithm(name :str, force: bool=False, async_upload: bool=False):
    """
    Create/update and/or upload local algorithm to remote Anylearn.
    """
    config = AnylearnCliConfig.load()
    try:
        algo = config.algorithms[name]
        dir = config.path['algorithm'][name]
        img = config.images[name]
    except KeyError:
        cmd_error(msg=(
            f"Algorithm named {name} or its path config does not exist."
        ))
        raise click.Abort
    uploader = AsyncResourceUploader() if async_upload else SyncResourceUploader()
    algo = _push_1_algorithm(
        algorithm=algo,
        dir=dir,
        image=img,
        force=force,
    )
    config.algorithms[name] = algo
    AnylearnCliConfig.update(config)
    cmd_success(msg="PUSHED")


def _push_1_algorithm(
    algorithm: Algorithm,
    dir: Optional[str]=None,
    image: Optional[str]='QUICKSTART',
    force: bool=False,
    uploader=None,
    polling=5,
) -> Algorithm:
    if dir:
        return __push_local_algorithm(
            algorithm=algorithm,
            dir=dir,
            image=image,
            force=force,
        )
    else:
        return __push_remote_algorithm(
            algorithm=algorithm,
            force=force,
        )


def __push_local_algorithm(
    algorithm: Algorithm,
    dir: str,
    image: str,
    force: bool=False,
):
    try:
        algo, _ = sync_algorithm(
            name=algorithm.name,
            dir_path=dir,
            entrypoint_training=algorithm.entrypoint_training,
            output_training=algorithm.output_training,
            mirror_name=image,
            force=False,
        )
        return algo
    except AnylearnRequiredLocalCommitException as e:
        cmd_warning(msg=(
            f"Algorithm dir {dir} is not clean, commit required. "
            "Anylearn can make an auto-commit in this case "
            "(or you can cancel the operation and "
            "commit your changes yourself)."
        ))
        if not force:
            cmd_confirm_or_abort()
        algo, _ = sync_algorithm(
            name=algorithm.name,
            dir_path=dir,
            entrypoint_training=algorithm.entrypoint_training,
            output_training=algorithm.output_training,
            mirror_name=image,
            force=True,
        )
        return algo


def __push_remote_algorithm(algorithm: Algorithm, force: bool=False):
    try:
        remote_algo = Algorithm(id=algorithm.id, load_detail=True)
        if remote_algo == algorithm:
            cmd_info(msg="Metadata already up-to-date.")
        else:
            cmd_warning(msg=(
                "Remote algorithm "
                "("
                f"id={remote_algo.id}, "
                f"name={remote_algo.name}"
                ") will be overridden."
            ))
            if not force:
                cmd_confirm_or_abort()
            algorithm.save()
    except HTTPError:
        cmd_error(msg=(
            "Remote algorithm "
            "("
            f"id={algorithm.id}, "
            f"name={algorithm.name}"
            ") is unaccessible."
        ))
        raise click.Abort
    return algorithm


@commands.command()
@click.argument('name')
@_option_force
@_option_async_upload
@check_config()
@check_connection()
@get_cmd_command()
def dataset(name :str, force: bool=False, async_upload: bool=False):
    """
    Create/update and/or upload local dataset to remote Anylearn.
    """
    config = AnylearnCliConfig.load()
    try:
        dset = config.datasets[name]
        dir = config.path['dataset'][name]
    except KeyError:
        cmd_error(msg=(
            f"Dataset named {name} or its path config does not exist."
        ))
        raise click.Abort
    uploader = AsyncResourceUploader() if async_upload else SyncResourceUploader()
    dset = _push_1_dataset(
        dataset=dset,
        dir=dir,
        force=force,
        uploader=uploader,
    )
    if dset:
        config.datasets[name] = dset
        AnylearnCliConfig.update(config)
        cmd_success(msg="PUSHED")


def _push_1_dataset(
    dataset: Dataset,
    dir: Optional[str]=None,
    force: bool=False,
    uploader=None,
    polling=5,
) -> Dataset:
    if dataset.id:
        return __push_remote_dataset(
            dataset=dataset,
            force=force,
        )
    else:
        return __push_local_dataset(
            dataset=dataset,
            dir=dir,
            uploader=uploader,
            polling=polling,
        )


def __push_local_dataset(
    dataset: Dataset,
    dir: Optional[str]=None,
    uploader=None,
    polling=5,
):
    return sync_dataset(
        id=dataset.id,
        name=dataset.name,
        dir_path=dir,
        uploader=uploader,
        polling=polling,
    )


def __push_remote_dataset(dataset: Dataset, force: bool=False):
    try:
        remote_dset = Dataset(id=dataset.id, load_detail=True)
        if remote_dset == dataset:
            cmd_info(msg="Metadata already up-to-date.")
        else:
            cmd_warning(msg=(
                "Remote dataset "
                "("
                f"id={remote_dset.id}, "
                f"name={remote_dset.name}"
                ") will be overridden."
            ))
            if not force:
                cmd_confirm_or_abort()
            dataset.save()
    except HTTPError:
        cmd_error(msg=(
            "Remote dataset "
            "("
            f"id={dataset.id}, "
            f"name={dataset.name}"
            ") is unaccessible."
        ))
        raise click.Abort
    return dataset



from dataclasses import asdict, dataclass, field
from datetime import datetime
import json
from typing import Optional

from anylearn.utils.api import url_base, get_with_token, post_with_token
from anylearn.utils.errors import AnyLearnException, AnyLearnMissingParamException
from anylearn.interfaces.base import BaseObject


class EvaluateTaskState:
    """
    训练任务状态标识：

    - 0(CREATED)表示已创建
    - 1(RUNNING)表示运行中
    - 2(SUCCESS)表示已完成
    - -1(DELETED)表示已删除
    - -2(FAIL)表示失败
    - -3(ABORT)表示中断
    """
    CREATED = 0
    RUNNING = 1
    SUCCESS = 2
    DELETED = -1
    FAIL = -2
    ABORT = -3


class EvaluateTaskVisibility:
    """
    项目可见性标识：

    - -1(HIDDEN)表示不可见
    - 1(PRIVATE)表示仅创建者可见
    - 2(PROTECTED)表示所有者可见
    - 3(PUBLIC)表示公开
    """
    HIDDEN = -1
    PRIVATE = 1
    PROTECTED = 2
    PUBLIC = 3


@dataclass
class EvaluateSubTask:
    """
    AnyLearn验证子任务类，以方法映射数据集CRUD相关接口

    Attributes
    ----------
    id
        子任务ID，格式为EVRE+uuid1生成的编码中后28个有效位（小写字母和数字）组成
    model_id
        模型ID
    files
        数据集ID
    evaluate_params
        验证参数，包括 :obj:`dataset` 、 :obj:`eval_metric` 、 :obj:`model_path`
    envs
        环境变量
    gpu_num
        申请的GPU数目
    results
        验证结果
    """

    id: Optional[str]=None
    model_id: Optional[str]=None
    files: list = field(default_factory=list)
    evaluate_params: dict = field(default_factory=dict)
    envs: dict = field(default_factory=dict)
    gpu_num: int = 1
    results: Optional[str]=None


class EvaluateTask(BaseObject):
    """
    AnyLearn验证项目类，以方法映射数据集CRUD相关接口

    Attributes
    ----------
    id
        验证任务的唯一标识符，自动生成，由EVAL+uuid1生成的编码中后28个有效位（小写字母和数字）组成
    name
        验证任务的名称
    description
        验证任务描述
    state
        验证任务状态
    visibility
        验证任务的可见性
    creator_id
        验证任务的创建者
    owner
        验证任务的所有者，以逗号分隔的这些用户的ID拼成的字符串，无多余空格
    create_time
        验证任务创建时间
    finish_time
        验证任务完成时间
    params_list
        验证任务参数列表
    sub_tasks
        子验证任务
    secret_key
        秘钥
    load_detail
        初始化时是否加载详情
    """

    _fields = {
        # 资源创建/更新请求包体中必须包含且不能为空的字段
        'required': {
            'create': ['name', 'sub_tasks'],
            'update': ['id', 'name'],
        },
        # 资源创建/更新请求包体中包含的所有字段
        'payload': {
            'create': ['id', 'name', 'description', 'visibility', 'owner',
                       'params_list'],
            'update': ['id', 'name', 'description', 'visibility', 'owner'],
        },
    }
    """
    创建/更新对象时：

    - 必须包含且不能为空的字段 :obj:`_fields['required']`
    - 所有字段 :obj:`_fields['payload']`
    """

    def __init__(self,
                 id: Optional[str]=None,
                 name: Optional[str]=None,
                 description: Optional[str]=None,
                 state: Optional[int]=None,
                 visibility: Optional[int]=None,
                 creator_id: Optional[str]=None,
                 owner: Optional[list]=None,
                 create_time: Optional[datetime]=None,
                 finish_time: Optional[datetime]=None,
                 params_list: Optional[str]=None,
                 sub_tasks: Optional[list]=None,
                 secret_key: Optional[str]=None,
                 load_detail=False):
        """
        Parameters
        ----------
        id
            验证任务的唯一标识符，自动生成，由EVAL+uuid1生成的编码中后28个有效位（小写字母和数字）组成
        name
            验证任务的名称
        description
            验证任务描述
        state
            验证任务状态
        visibility
            验证任务的可见性
        creator_id
            验证任务的创建者
        owner
            验证任务的所有者，以逗号分隔的这些用户的ID拼成的字符串，无多余空格
        create_time
            验证任务创建时间
        finish_time
            验证任务完成时间
        params_list
            验证任务参数列表
        sub_tasks
            子验证任务
        secret_key
            秘钥
        load_detail
            初始化时是否加载详情
        """
        self.id = id
        self.name = name
        self.description = description
        self.state = state
        self.visibility = visibility
        self.creator_id = creator_id
        self.owner = owner
        self.create_time = create_time
        self.finish_time = finish_time
        self.params_list = params_list
        self.sub_tasks = sub_tasks
        self.secret_key = secret_key
        super().__init__(id, load_detail=load_detail)

    @classmethod
    def get_list(cls):
        """
        获取验证任务列表
        
        Returns
        -------
        List [EvaluateTask]
            验证任务对象的集合。
        """
        res = get_with_token(f"{url_base()}/evaluate_task/list")
        if res is None or not isinstance(res, list):
            raise AnyLearnException("请求未能得到有效响应")
        return [
            EvaluateTask(id=e['id'], name=e['name'],
                         description=e['description'], state=e['state'],
                         visibility=e['visibility'],
                         creator_id=e['creator_id'], owner=e['owner'],
                         create_time=e['create_time'],
                         finish_time=e['finish_time'],
                         secret_key=e['secret_key'])
            for e in res
        ]

    def get_detail(self):
        """
        获取验证任务详细信息

        - 对象属性 :obj:`id` 应为非空

        Returns
        -------
        EvaluateTask
            验证任务对象。
        """
        self._check_fields(required=['id'])
        res = get_with_token(f"{url_base()}/evaluate_task/query",
                             params={'id': self.id})
        if not res or not isinstance(res, dict):
            raise AnyLearnException("请求未能得到有效响应")
        self.__init__(id=res['id'], name=res['name'],
                      description=res['description'], state=res['state'],
                      visibility=res['visibility'],
                      creator_id=res['creator_id'], owner=res['owner'],
                      create_time=res['create_time'],
                      finish_time=res['finish_time'],
                      params_list=res['params_list'],
                      secret_key=res['secret_key'])
        if 'results' not in res or not res['results']:
            self.__load_sub_tasks_from_params_list()
        else:
            self.__load_sub_tasks_from_results(res['results'])

    def _create(self):
        self.__dump_sub_tasks_to_params_list()
        return super()._create()

    def get_log(self, limit=100, direction="init", offset=0, includes=False):
        """
        验证任务日志查询接口

        - 对象属性 :obj:`id` 应为非空
        
        :param limit: :obj:`int`
                    日志条数上限（默认值100）。
        :param direction: :obj:`str`
                    日志查询方向。
        :param offset: :obj:`int`
                    日志查询索引。
        :param includes: :obj:`bool`
                    是否包含指定索引记录本身。

        :return: 
            .. code-block:: json

                {
                  "EVRE123": [
                    {
                      "offset": 2955,
                      "text": "Task EVRE123 finished."
                    }
                  ]
                }
        """
        self._check_fields(required=['id'])
        params = {
            'id': self.id,
            'limit': limit,
            'direction': direction,
            'index': offset,
            'self': 1 if includes else 0,
        }
        res = get_with_token(f"{url_base()}/evaluate_task/log", params=params)
        if not res or type(res) != dict:
            raise AnyLearnException("请求未能得到有效响应")
        return res

    def _namespace(self):
        return "evaluate_task"
    
    def __dump_sub_tasks_to_params_list(self):
        sub_tasks = [asdict(t) for t in self.sub_tasks]
        for t in sub_tasks:
            if not t['model_id']:
                raise AnyLearnMissingParamException("EvaluateSubTask缺少必要字段：model_id")
            t['files'] = ",".join(t['files'])
            t['envs'] = ",".join([f"{k}={v}" for k, v in t['envs'].items()])
        self.params_list = json.dumps(sub_tasks)
        return self.params_list

    def __load_sub_tasks_from_params_list(self):
        if not self.params_list:
            self.sub_tasks = []
            return []
        sub_tasks = json.loads(self.params_list)
        self.sub_tasks = [
            EvaluateSubTask(id=t['id'] if 'id' in t else None,
                            model_id=t['model_id'],
                            files=t['files'].split(",") if t['files'] else [],
                            evaluate_params=t['evaluate_params'],
                            envs={kv[0]: kv[1]
                                    for kv in [
                                        env.split("=")
                                        for env in t['envs'].split(",")
                                    ]} if 'envs' in t and t['envs'] else {},
                            gpu_num=t['gpu_num'] if 'gpu_num' in t else 1,
                            results=t['results'] if 'results' in t else None)
            for t in sub_tasks
        ]
        return self.sub_tasks

    def __load_sub_tasks_from_results(self, results: list):
        self.sub_tasks = [
            EvaluateSubTask(id=t['id'],
                            model_id=t['model_id'],
                            files=t['files'].split(","),
                            evaluate_params=json.loads(t['args']),
                            envs={kv[0]: kv[1]
                                    for kv in [
                                        env.split("=")
                                        for env in t['envs'].split(",")
                                    ]} if t['envs'] else {},
                            gpu_num=t['gpu_num'], results=t['results'])
            for t in results
        ]
        return self.sub_tasks



import os
import aiohttp
import responses
from aioresponses import aioresponses
from requests import RequestException

from anylearn.interfaces.resource.resource_downloader import AsyncResourceDownloader, SyncResourceDownloader
from tests.base_test_case import BaseTestCase

class TestResourceDownloader(BaseTestCase):
    @responses.activate
    def test_run_async_resource_downloader_ok(self):
        downloader = AsyncResourceDownloader()
        responses.add(responses.POST, url=self._url("resource/compression"),
                        match_querystring=True,
                        json={'data': "COMP123"},
                        status=200)
        responses.add(responses.GET, url=self._url("resource/compression"),
                            match_querystring=True,
                            json={'data': True},
                            status=200)
        with aioresponses() as mock_responses:
            file_name = "test_async_download.zip"
            headers = {'Content-Disposition': "attachment; filename=%s;" % file_name,}
            mock_responses.get(url=self._url("resource/download?file_id=FILE001&compression_id=COMP123&token=TEST_TOKEN"),
                               headers=headers,
                               payload={
                                   'message': "OK",
                               },
                               status=200)
            res = downloader.run(resource_id="FILE001",
                                 save_path="tests/",
                                 polling=1)
        os.remove(f"tests/{file_name}")
        self.assertTrue(res)

    @responses.activate
    def test_run_async_resource_downloader_500(self):
        with aioresponses() as mock_responses:
            downloader = AsyncResourceDownloader()
            responses.add(responses.POST, url=self._url("resource/compression"),
                          match_querystring=True,
                          json={'data': "COMP123"},
                          status=200)
            responses.add(responses.GET, url=self._url("resource/compression"),
                               match_querystring=True,
                               json={'data': True},
                               status=200)
            mock_responses.get(url=self._url("resource/download?file_id=FILE001&compression_id=COMP123&token=TEST_TOKEN"),
                               status=500)
            with self.assertRaises(aiohttp.ClientResponseError) as ctx:
                downloader.run(resource_id="FILE001",
                               save_path="/data/wtt",
                               polling=1)
                e = ctx.exception
                self.assertIsInstance(e,aiohttp.ClientResponseError)

    @responses.activate
    def test_run_sync_resource_downloader_ok(self):
        responses.add(responses.POST, url=self._url("resource/compression"),
                        match_querystring=True,
                        json={'data': "COMP123"},
                        status=200)
        responses.add(responses.GET, url=self._url("resource/compression"),
                            match_querystring=True,
                            json={'data': True},
                            status=200)
        downloader = SyncResourceDownloader()
        file_name = "test_sync_download.zip"
        headers = {'Content-Disposition': "attachment; filename=%s;" % file_name,}
        responses.add(responses.GET, url=self._url("resource/download?file_id=FILE001&compression_id=COMP123&token=TEST_TOKEN"),
                      match_querystring=True,
                      headers=headers,
                      status=200)
        res = downloader.run(resource_id="FILE001",
                             save_path="tests/")
        os.remove(f"tests/{file_name}")
        self.assertTrue(res)

    @responses.activate
    def test_run_sync_resource_downloader_500(self):
        downloader = SyncResourceDownloader()
        responses.add(responses.GET, url=self._url("resource/download?file_id=FILE001&token=TEST_TOKEN"),
                      status=500)
        with self.assertRaises(RequestException) as ctx:
            downloader.run(resource_id="FILE001",
                           save_path="/data/wtt")
        e = ctx.exception
        self.assertIsInstance(e, RequestException)



import json
from pathlib import Path
import time

import click

from anylearn.applications.hpo import run_hpo, run_hpo_trial
from anylearn.applications.quickstart import quick_train
from anylearn.config import AnylearnConfig
from anylearn.utils.errors import AnyLearnException, AnyLearnMissingParamException


@click.group()
@click.version_option()
def cli():
    pass


@cli.command()
@click.argument("cluster_address")
# TODO: implement API token auth and replace the basic auth below
@click.argument("username")
@click.argument("password")
@click.argument("search_space_file")
@click.option(
    "-f", "--foreground",
    is_flag=True,
    default=False,
    help="""Run HPO in foreground mode with log print in terminal.
    """
)
@click.option(
    "--algorithm-dir",
    default=None,
    help="""Local algorithm folder absolute path.
    """
)
@click.option(
    "--algorithm-entrypoint",
    default=None,
    help="""Entrypoint command of algorithm. 
    Required when using local algorithm.
    """
)
@click.option(
    "--algorithm-output",
    default=None,
    help="""Model saving path of algorithm. 
    Required when using local algorithm.
    """
)
@click.option(
    "--dataset-dir",
    default=None,
    help="""Local dataset folder absolute path.
    """
)
@click.option(
    "--project-name",
    default=None,
    help="""Name of project to assemble all HPO trials.
    """
)
@click.option(
    "--dataset-hyperparam-name",
    default="dataset",
    help="""Parameter name to pass dataset into algorithm.
    """
)
@click.option(
    "--gpu-num",
    default=1,
    help="""Number of GPU claimed for trial.
    """
)
@click.option(
    "--hpo-max-runs",
    default=10,
    help="""Max number of trials to run in HPO
    """
)
@click.option(
    "--hpo-max-duration",
    default="24h",
    help="""Max duration of HPO, string finished by time unit: 's', 'm', 'h'.
    """
)
@click.option(
    "--hpo-tuner-name",
    default="TPE",
    help="""Name of HPO tuner supported by NNI.
    """
)
@click.option(
    "--hpo-mode",
    default="maximize",
    help="""Optimization mode either 'maximize' or 'minimize'.
    """
)
@click.option(
    "--hpo-port",
    default=31889,
    help="""Port to bind HPO process (NNI) and hence its web view.
    """
)
def hpo(
    cluster_address,
    username,
    password,
    foreground,
    algorithm_dir,
    dataset_dir,
    algorithm_entrypoint,
    algorithm_output,
    search_space_file,
    project_name,
    dataset_hyperparam_name,
    gpu_num,
    hpo_max_runs,
    hpo_max_duration,
    hpo_tuner_name,
    hpo_mode,
    hpo_port,
):
    with open(search_space_file) as f:
        hpo_search_space = json.loads(f.read())
    AnylearnConfig.init(
        cluster_address=cluster_address,
        username=username,
        password=password,
    )

    print(
        f"""---
        Algorithm folder: {algorithm_dir}
        Dataset folder: {dataset_dir}
        ---
        Starting HPO...
        """
    )
    # Note that only foreground mode works for now,
    # since NNI's REST server runs in subprocess
    # which would be killed if anylearnctl does not
    # hang in terminal.
    # TODO: make run_hpo work in process
    run_hpo(
        hpo_search_space=hpo_search_space,
        hpo_max_runs=hpo_max_runs,
        hpo_max_duration=hpo_max_duration,
        hpo_tuner_name=hpo_tuner_name,
        hpo_mode=hpo_mode,
        hpo_port=hpo_port,
        algorithm_dir=algorithm_dir,
        algorithm_entrypoint=algorithm_entrypoint,
        algorithm_output=algorithm_output,
        dataset_dir=dataset_dir,
        project_name=project_name,
        dataset_hyperparam_name=dataset_hyperparam_name,
        gpu_num=gpu_num,
        foreground=foreground,
    )


@cli.command()
@click.argument("cluster_address")
# TODO: implement API token auth and replace the basic auth below
@click.argument("username")
@click.argument("password")
@click.argument("project_id")
@click.argument("algorithm_id")
@click.argument("dataset_id")
@click.option(
    "--dataset-hyperparam-name",
    default="dataset",
    help="""Parameter name to pass dataset into algorithm.
    """
)
@click.option(
    "--hyperparam",
    multiple=True,
    help="""Hyperparameters for trial formated '<key>=<value>'. 
    This option could be used multiple times to pass multiple hyperparameters.
    \b\r\n
    Example:
    anylearnctl trial --hyperparam batch_size=64 --hyperparam lr=0.01 [OTHER_OPTIONS]
    """
)
@click.option(
    "--gpu-num",
    default=1,
    help="""Number of GPU claimed for trial.
    """
)
def trial(
    cluster_address,
    username,
    password,
    project_id,
    algorithm_id,
    dataset_id,
    dataset_hyperparam_name,
    hyperparam,
    gpu_num,
):
    AnylearnConfig.init(
        cluster_address=cluster_address,
        username=username,
        password=password,
    )
    print(
        f"""---
        Project ID: {project_id}
        Algorithm ID: {algorithm_id}
        Dataset ID: {dataset_id}
        ---
        Running trial...
        """
    )
    hyperparams = {k: v for k, v in [
        (i.split("=")[0], i.split("=")[1])
        for i in list(hyperparam)
    ]}
    train_task = run_hpo_trial(
        project_id=project_id,
        algorithm_id=algorithm_id,
        dataset_id=dataset_id,
        dataset_hyperparam_name=dataset_hyperparam_name,
        hyperparams=hyperparams,
        gpu_num=gpu_num,
    )
    print("---")
    metric = train_task.get_final_metric()
    print(f"Final metric: {metric}")


MSG_HELP_TRAIN_ALGO_OPTIONS = (
    "At least one of the options "
    "["
    "'--algorithm-id', "
    "'--algorithm-dir', "
    "'--algorithm-archive'"
    "] "
    "should be specified."
)
MSG_HELP_TRAIN_DSET_OPTIONS = (
    "At least one of the options "
    "["
    "'--dataset-id', "
    "'--dataset-dir', "
    "'--dataset-archive'"
    "] "
    "should be specified."
)


@cli.command()
@click.argument("cluster_address")
# TODO: implement API token auth and replace the basic auth below
@click.argument("username")
@click.argument("password")
@click.option(
    "-d", "--detach",
    is_flag=True,
    default=False,
    help="""Launch training and print remote training task ID 
    without waiting training result.
    """
)
@click.option(
    "--algorithm-id",
    default=None,
    help="""Remote algorithm ID.
    \b\r\n
    """ + MSG_HELP_TRAIN_ALGO_OPTIONS,
)
@click.option(
    "--algorithm-dir",
    default=None,
    help="""Local algorithm folder absolute path.
    \b\r\n
    """ + MSG_HELP_TRAIN_ALGO_OPTIONS,
)
@click.option(
    "--algorithm-archive",
    default=None,
    help="""Local algorithm archive absolute path.
    \b\r\n
    """ + MSG_HELP_TRAIN_ALGO_OPTIONS,
)
@click.option(
    "--dataset-id",
    default=None,
    help="""Remote dataset ID.
    \b\r\n
    """ + MSG_HELP_TRAIN_DSET_OPTIONS,
)
@click.option(
    "--dataset-dir",
    default=None,
    help="""Local dataset folder absolute path.
    \b\r\n
    """ + MSG_HELP_TRAIN_DSET_OPTIONS,
)
@click.option(
    "--dataset-archive",
    default=None,
    help="""Local dataset archive absolute path.
    \b\r\n
    """ + MSG_HELP_TRAIN_DSET_OPTIONS,
)
@click.option(
    "--project-id",
    default=None,
    help="""Remote training project ID.
    """
)
@click.option(
    "--algorithm-entrypoint",
    default=None,
    help="""Entrypoint command of algorithm. 
    Required when using local algorithm.
    """
)
@click.option(
    "--algorithm-output",
    default=None,
    help="""Model saving path of algorithm. 
    Required when using local algorithm.
    """
)
@click.option(
    "--dataset-hyperparam-name",
    default="dataset",
    help="""Parameter name to pass dataset into algorithm.
    """
)
@click.option(
    "--hyperparam",
    multiple=True,
    help="""Hyperparameters for training formated '<key>=<value>'. 
    This option could be used multiple times to pass multiple hyperparameters.
    \b\r\n
    Example:
    anylearnctl train --hyperparam batch_size=64 --hyperparam lr=0.01 [OTHER_OPTIONS]
    """
)
@click.option(
    "--gpu-num",
    default=1,
    help="""Number of GPU claimed for training.
    """
)
def train(
    cluster_address,
    username,
    password,
    detach,
    algorithm_id,
    algorithm_dir,
    algorithm_archive,
    dataset_id,
    dataset_dir,
    dataset_archive,
    project_id,
    algorithm_entrypoint,
    algorithm_output,
    dataset_hyperparam_name,
    hyperparam,
    gpu_num,
):
    if not any([algorithm_id, algorithm_dir, algorithm_archive]):
        raise AnyLearnMissingParamException(MSG_HELP_TRAIN_ALGO_OPTIONS)
    if not any([dataset_id, dataset_dir, dataset_archive]):
        raise AnyLearnMissingParamException(MSG_HELP_TRAIN_DSET_OPTIONS)
    AnylearnConfig.init(
        cluster_address=cluster_address,
        username=username,
        password=password,
    )
    hyperparams = {k: v for k, v in [
        (i.split("=")[0], i.split("=")[1])
        for i in list(hyperparam)
    ]}
    train_task, algo, dset, project = quick_train(
        algorithm_id=algorithm_id,
        algorithm_dir=algorithm_dir,
        algorithm_archive=algorithm_archive,
        dataset_id=dataset_id,
        dataset_dir=dataset_dir,
        dataset_archive=dataset_archive,
        project_id=project_id,
        entrypoint=algorithm_entrypoint,
        output=algorithm_output,
        dataset_hyperparam_name=dataset_hyperparam_name,
        hyperparams=hyperparams,
        gpu_num=gpu_num,
    )
    print(
        f"""---
        Project ID: {project.id}
        Training ID: {train_task.id}
        Algorithm ID: {algo.id}
        Dataset ID: {dset.id}
        """
    )
    if detach:
        return
    print(
        """---
        Waiting training result...
        """
    )
    while not train_task.finished():
        time.sleep(20) # TODO: make polling interval dynamic
        train_task.get_detail()
        try:
            metric = train_task.get_intermediate_metric()
            print(f"- Intermediate metric: {metric}")
        except:
            continue
    print("---")
    print(f"Final metric: {train_task.get_final_metric()}")
    print(f"Model export from CLI will soon be supported.")


@cli.command()
def evaluate():
    print("Coming soon.")


@cli.command()
def serve():
    print("Coming soon.")


@cli.command()
def resources():
    print("Coming soon.")


if __name__ == '__main__':
    cli()



import os
from pathlib import Path
import requests
import shutil

from anylearn.utils import logger
from anylearn.utils.errors import InvalidArguments, AnyLearnAuthException


class AnylearnConfig(object):
    """
    Anylearn SDK配置类，
    包含Anylearn后端引擎的连接配置和SDK的本地存储配置。
    """

    cluster_address = None
    """Anylearn后端引擎集群网关地址"""

    username = None
    """Anylearn后端引擎账户用户名"""

    user_id = None
    """Anylearn后端引擎账户ID"""

    password = None
    """Anylearn后端引擎账户密码"""

    token = None
    """Anylearn后端引擎API令牌"""

    workspace_path = Path().home() / ".anylearn"
    """SDK本地存储工作区路径"""

    db_path = workspace_path / "anylearn.db"
    """SDK本地存储元信息数据库文件（ :obj:`sqlite`）路径"""

    _db_scheme = "sqlite:///" if os.name == 'nt' else "sqlite:////"

    db_uri = f"{_db_scheme}{db_path.absolute()}"
    """SDK本地存储元信息数据库（ :obj:`sqlite`）的URI"""

    git_ready = False

    @classmethod
    def init(cls, cluster_address=None, username=None, password=None,
             token=None, workspace=None, disable_git=False):
        """
        初始化SDK与后端连接以及本地存储。
        可通过用户名密码建立后端连接，也可直接传入有效的后端API令牌。
        
        Parameters
        ----------
        cluster_address : :obj:`str`, optional
            Anylearn后端引擎集群网关地址。
        username : :obj:`str`, optional
            Anylearn后端引擎账户用户名。
        password : :obj:`str`, optional
            Anylearn后端引擎账户密码。
        token : :obj:`str`, optional
            Anylearn后端引擎API令牌。
        workspace : :obj:`str`, optional
            SDK本地存储工作区路径，
            默认将工作区建立在 :obj:`<home>/.anylearn` 目录下。
        """
        # Cluster Auth
        cls.init_cluster(cluster_address, username, password, token)
        # Local workspace
        cls.init_workspace(workspace)
        # Detect git
        try:
            from git import Repo
            cls.git_ready = not disable_git
        except:
            logger.warning("Git executable not found.")
            cls.git_ready = False
        
    @classmethod
    def init_cluster(cls, cluster_address=None, username=None, password=None,
                     token=None):
        """
        仅初始化SDK与后端连接。
        可通过用户名密码建立后端连接，也可直接传入有效的后端API令牌。
        
        Parameters
        ----------
        cluster_address : :obj:`str`, optional
            Anylearn后端引擎集群网关地址。
        username : :obj:`str`, optional
            Anylearn后端引擎账户用户名。
        password : :obj:`str`, optional
            Anylearn后端引擎账户密码。
        token : :obj:`str`, optional
            Anylearn后端引擎API令牌。
        """
        if not all([
            cluster_address,
            isinstance(cluster_address, str),
            any([username and password, token]),
        ]):
            raise InvalidArguments('无有效登录配置')
        if cluster_address.endswith('/'):
            cluster_address = cluster_address[:-1]
        cls.cluster_address = cluster_address
        cls.username = username
        cls.password = password
        cls.token = token
        if token is None:
            cls.cluster_login()
    
    @classmethod
    def cluster_login(cls):
        """
        通过用户名密码建立SDK与后端连接，
        获取后端API令牌。
        在以用户名密码的方式初始化后方可调用。
        """
        res = requests.post('%s/api/user/login' % cls.cluster_address, {
            'username': cls.username,
            'password': cls.password,
        })
        if res.status_code == 200:
            res = res.json()
            cls.token = res['token']
            cls.user_id = res['id']
            cls.username = res['username']
        else:
            raise AnyLearnAuthException()
    
    @classmethod
    def init_workspace(cls, workspace=None):
        """
        仅初始化SDK本地存储。
        
        Parameters
        ----------
        workspace : :obj:`str`, optional
            SDK本地存储工作区路径，
            默认将工作区建立在 :obj:`<home>/.anylearn` 目录下。
        """
        cls.workspace_path = Path(workspace) if workspace \
                                             else cls.workspace_path
        cls.db_path = cls.workspace_path / "anylearn.db"
        cls.db_uri = f"{cls._db_scheme}{cls.db_path}"
        
        # Remove if workspace path points to a file
        if cls.workspace_path.is_file():
            cls.workspace_path.unlink()
        
        # Ensure workspace folder exists
        cls.workspace_path.mkdir(exist_ok=True)
        
        # Ensure DB file exists
        if not cls.db_path.exists():
            cls.db_path.touch()

    @classmethod
    def clear_workspace(cls):
        """
        清空SDK本地存储。
        本地元信息存储、临时文件、资源缓存等均会被删除。
        请谨慎调用。
        """
        shutil.rmtree(cls.workspace_path)


def init_sdk(cluster_address, username, password, disable_git=False):
    """
    初始化SDK与后端连接的接口。
    调用本接口并传入后端地址、用户名和密码，
    SDK将自动以相应账户进行登录并获取后端API令牌。

    Parameters
    ----------
    cluster_address : :obj:`str`
        Anylearn后端引擎集群网关地址。
    username : :obj:`str`
        Anylearn后端引擎账户用户名。
    password : :obj:`str`
        Anylearn后端引擎账户密码。
    """
    AnylearnConfig.init(
        cluster_address=cluster_address,
        username=username,
        password=password,
        disable_git=disable_git,
    )


def init_sdk_incontainer(cluster_address):
    """
    初始化无鉴权信息的SDK与后端连接的接口。
    调用本接口仅需传入后端地址，
    无需账户信息。
    以此方式初始化的SDK将无法调用需要鉴权信息的接口，
    仅可调用无需API令牌的公共接口。

    此接口的使用场景多为任务执行容器内与后端建立通信，
    不建议用户使用。

    Parameters
    ----------
    cluster_address : :obj:`str`
        Anylearn后端引擎集群网关地址。
    """
    AnylearnConfig.init(
        cluster_address=cluster_address,
        token="INCONTAINER"
    )


def print_config():
    print(AnylearnConfig.cluster_address)



from datetime import datetime
import requests
from typing import Optional

from anylearn.utils.api import url_base, get_with_token, put_with_token, patch_with_token
from anylearn.utils.errors import AnyLearnException
from anylearn.interfaces.base import BaseObject


class User(BaseObject):
    """
    AnyLearn用户类，以方法映射训练任务CRUD相关接口
    
    Attributes
    ----------
    id
        用户的唯一标识符，自动生成，由USER+uuid1生成的编码中后28个有效位（小写字母和数字）组成
    token
        单点登录令牌
    username
        用户名
    password
        用户密码
    namespace
        命名空间
    email
        邮箱
    role
        用户角色
    algorithm_tree
        暂时忽略不用
    collected_algorithm
        暂时忽略不用
    own_algorithms
        用户算法集合
    own_models
        用户模型集合
    own_files
        用户文件集合
    create_time
        创建时间
    own_datasets
        用户数据集集合
    load_detail
        初始化时是否加载详情
    """

    _fields = {
        # 资源创建/更新请求包体中必须包含且不能为空的字段
        'required': {
            'create': ['username', 'password', 'email'],
            'update': ['id'],
        },
        # 资源创建/更新请求包体中包含的所有字段
        'payload': {
            'create': ['id', 'username', 'password', 'email'],
            'update': ['id', 'username', 'email', 'role'],
        },
    }
    """
    创建/更新对象时：

    - 必须包含且不能为空的字段 :obj:`_fields['required']`
    - 所有字段 :obj:`_fields['payload']`
    """

    def __init__(self,
                 id: Optional[str]=None,
                 token: Optional[str]=None,
                 username: Optional[str]=None,
                 password: Optional[str]=None,
                 namespace: Optional[str]=None,
                 email: Optional[str]=None,
                 role: Optional[str]=None,
                 algorithm_tree: Optional[str]=None,
                 collected_algorithm: Optional[list]=None,
                 own_algorithms: Optional[list]=None,
                 own_models: Optional[list]=None,
                 own_files: Optional[list]=None,
                 create_time: Optional[datetime]=None,
                 own_datasets: Optional[list]=None,
                 load_detail=False):
        """
        Parameters
        ----------
        id
            用户的唯一标识符，自动生成，由USER+uuid1生成的编码中后28个有效位（小写字母和数字）组成
        token
            单点登录令牌
        username
            用户名
        password
            用户密码
        namespace
            命名空间
        email
            邮箱
        role
            用户角色
        algorithm_tree
            暂时忽略不用
        collected_algorithm
            暂时忽略不用
        own_algorithms
            用户算法集合
        own_models
            用户模型集合
        own_files
            用户文件集合
        create_time
            创建时间
        own_datasets
            用户数据集集合
        load_detail
            初始化时是否加载详情
        """
        self.id = id
        self.token = token
        self.username = username
        self.password = password
        self.namespace = namespace
        self.email = email
        self.role = role
        self.algorithm_tree = algorithm_tree
        self.collected_algorithm = collected_algorithm
        self.own_algorithms = own_algorithms
        self.own_datasets = own_datasets
        self.own_models = own_models
        self.own_files = own_files
        self.create_time = create_time
        self.load_detail = load_detail
        super().__init__(id=id, load_detail=load_detail)

    def _url_create(self):
        return f"{url_base()}/{self._namespace()}/registry"

    def _url_update(self):
        return f"{url_base()}/{self._namespace()}"

    def _url_delete(self):
        return f"{url_base()}/{self._namespace()}"

    def _create(self):
        data = self._payload_create()
        res = requests.post(self._url_create(), data=data)
        res.raise_for_status()
        res.encoding = "utf-8"
        res = res.json()
        if not res or 'data' not in res:
            raise AnyLearnException("请求未能得到有效响应")
        self.id = res['data']
        return True

    def get_detail(self):
        """
        获取用户详细信息

        - 对象属性 :obj:`id` 应为非空

        Returns
        -------
        User
            用户对象。
        """
        self._check_fields(required=['id'])
        res = get_with_token(f"{url_base()}/user",
                             params={'id': self.id})
        if not res or not isinstance(res, list):
            raise AnyLearnException("请求未能得到有效响应")
        res = res[0]
        self.__init__(id=res['id'], username=res['username'], namespace=res['namespace'],
                      email=res['email'], role=res['role'], own_algorithms=res['own_algorithms'],
                      collected_algorithm=res['collected_algorithm'], own_datasets=res['own_datasets'],
                      own_models=res['own_models'], own_files=res['own_files'])

    @classmethod
    def user_collection(cls, ids: str):
        """
        获取用户集合

        Parameters
        ----------
        ids : :obj:`str`
            要查询的用户ID,多用户用逗号 :obj:`,` 隔开。

        Returns
        -------
        List[User]
            用户对象的集合。
        """
        res = get_with_token(f"{url_base()}/user/collection",
                             params={'collection': ids})
        if res is None or not isinstance(res, list):
            raise AnyLearnException("请求未能得到有效响应")
        return [
            User(id=item['id'], username=item['username'],
                 role=item['role'], email=item['email'])
            for item in res
        ]

    def change_password(self, old_password: str, new_password: str):
        """
        修改密码

        - 对象属性 :obj:`id` 应为非空

        :param old_password: :obj:`str`
                        旧密码。
        :param new_password: :obj:`str`
                        新密码。

        :return: 
            .. code-block:: json

                {
                  "data": "USER123",
                  "message": "密码修改成功"
                }
        """

        self._check_fields(required=['id'])
        data = {
            'id': self.id,
            'old_password': old_password,
            'new_password': new_password,
        }
        res = put_with_token(f"{url_base()}/user/password",
                             data=data)
        if not res or 'data' not in res:
            raise AnyLearnException("请求未能得到有效响应")
        return res

    @classmethod
    def get_list(cls):
        """
        获取用户列表
        
        Returns
        -------
        List [User]
            用户对象的集合。
        """
        res = get_with_token(f"{url_base()}/user/list")
        if res is None or not isinstance(res, list):
            raise AnyLearnException("请求未能得到有效响应")
        return [
            User(id=item['id'], username=item['username'],
                 role=item['role'], email=item['email'])
            for item in res
        ]

    @classmethod
    def user_check(cls, username: str):
        """
        用户名查重

        Parameters
        ----------
        username : :obj:`str`
            检查用户名是否重复。
        
        Returns
        -------
        bool
            True or False
        """
        if not username:
            raise AnyLearnException("username值不能为空")
        res = requests.get(f"{url_base()}/user/check",
                           params={'username': username})
        res.raise_for_status()
        res.encoding = "utf-8"
        res = res.json()
        if not isinstance(res, bool):
            raise AnyLearnException("请求未能得到有效响应")
        return res

    @classmethod
    def update_role_resource_count(cls, max_gpu: int, max_mem: int):
        """
        更新角色GPU和显存最大数量

        :param max_gpu: :obj:`int`
                        GPU数量。
        :param max_mem: :obj:`int`
                        显存数量。

        :return: 
            .. code-block:: json

                {
                  "msg": "user角色最大GPU和显存数量已分别成功设置为2,26"
                }
        """
        data = {
            'max_gpu': max_gpu,
            'max_mem': max_mem,
        }
        res = put_with_token(f"{url_base()}/user/role",
                             data=data)
        if not res or str (max_mem) not in res['msg']:
            raise AnyLearnException("请求未能得到有效响应")
        return res

    @classmethod
    def update_user_resource_count(cls, id: str, max_gpu: int, max_mem: int):
        """
        单独更新某一用户GPU和显存最大数量

        :param id: :obj:`str`
                        被更新的用户ID。
        :param max_gpu: :obj:`int`
                        GPU数量。
        :param max_mem: :obj:`int`
                        显存数量。

        :return: 
            .. code-block:: json

                {
                  "data": "USER123",
                  "msg": "用户最大GPU和显存数量已分别成功设置为3,18"
                }
        """
        data = {
            'id': id,
            'max_gpu': max_gpu,
            'max_mem': max_mem,
        }
        res = patch_with_token(f"{url_base()}/user/user",
                             data=data)
        if not res or 'data' not in res:
            raise AnyLearnException("请求未能得到有效响应")
        return res

    def _namespace(self):
        return "user"



from datetime import datetime
import json
from os import path
import requests
from typing import Optional

from anylearn.config import AnylearnConfig
from anylearn.utils.api import url_base, get_with_token
from anylearn.utils.errors import AnyLearnException
from anylearn.interfaces.base import BaseObject
from .record import ServiceRecord, ServiceRecordState


class ServiceVisibility:
    """
    服务可见性标识：

    - 1(PRIVATE)表示仅创建者可见
    - 2(PROTECTED)表示所有者可见
    - 3(PUBLIC)表示公开
    """
    PRIVATE = 1
    PROTECTED = 2
    PUBLIC = 3


class ServiceState:
    """
    服务状态标识：

    - 0(CREATED)表示已创建
    - 1(RUNNING)表示已部署运行中
    - -1(DELETED)表示已删除
    - -2(STOPPED)表示已停止
    """
    CREATED = 0
    RUNNING = 1
    DELETED = -1
    STOPPED = -2


class Service(BaseObject):
    """
    Attributes
    ----------
    id
        模型服务的唯一标识符，自动生成，由SERV+uuid1生成的编码中后28个有效位（小写字母和数字）组成（非空）
    name
        模型服务的名称
    description
        模型服务描述（默认None，最大长度255）
    visibility
        模型服务的可见性（默认值3）
    model_id
        模型服务使用的模型ID
    address
        模型服务运行地址
    secret_key
        模型服务秘钥
    creator_id
        创建者ID
    envs
        环境变量
    replicas
        模型服务副本数量
    gpu_num
        模型服务申请的GPU数目
    gpu_mem
        模型服务申请的GPU显存
    state
        状态
    create_time
        创建时间
    owner
        模型服务的所有者，以逗号分隔的这些用户的ID拼成的字符串，无多余空格
    load_detail
        初始化时是否加载详情
    """
    _fields = {
        # 资源创建/更新请求包体中必须包含且不能为空的字段
        'required': {
            'create': ['name', 'model_id'],
            'update': ['id', 'name'],
        },
        # 资源创建/更新请求包体中包含的所有字段
        'payload': {
            'create': ['name', 'description', 'visibility', 'model_id', 'envs',
                       'replicas', 'gpu_num', 'gpu_mem', 'owner'],
            'update': ['id', 'name', 'description', 'visibility', 'owner'],
        },
    }
    """
    创建/更新对象时：

    - 必须包含且不能为空的字段 :obj:`_fields['required']`
    - 所有字段 :obj:`_fields['payload']`
    """

    def __init__(self,
                 id: Optional[str]=None,
                 name: Optional[str]=None,
                 description: Optional[str]=None,
                 visibility=ServiceVisibility.PUBLIC,
                 model_id: Optional[str]=None,
                 address: Optional[str]=None,
                 secret_key: Optional[str]=None,
                 creator_id: Optional[str]=None,
                 envs: Optional[str]=None,
                 replicas=1,
                 gpu_num=0,
                 gpu_mem=0,
                 state: Optional[int]=None,
                 create_time: Optional[datetime]=None,
                 owner: Optional[list]=None,
                 load_detail=False):
        """
        Parameters
        ----------
        id
            模型服务的唯一标识符，自动生成，由SERV+uuid1生成的编码中后28个有效位（小写字母和数字）组成（非空）
        name
            模型服务的名称
        description
            模型服务描述（默认None，最大长度255）
        visibility
            模型服务的可见性（默认值3）
        model_id
            模型服务使用的模型ID
        address
            模型服务运行地址
        secret_key
            模型服务秘钥
        creator_id
            创建者ID
        envs
            环境变量
        replicas
            模型服务副本数量
        gpu_num
            模型服务申请的GPU数目
        gpu_mem
            模型服务申请的GPU显存
        state
            状态
        create_time
            创建时间
        owner
            模型服务的所有者，以逗号分隔的这些用户的ID拼成的字符串，无多余空格
        load_detail
            初始化时是否加载详情
        """
        self.name = name
        self.description = description
        self.state = state
        self.visibility = visibility
        self.creator_id = creator_id
        self.owner = owner
        self.model_id = model_id
        self.address = address
        self.secret_key = secret_key
        self.envs = envs
        self.replicas = replicas
        self.gpu_num = gpu_num
        self.gpu_mem = gpu_mem
        self.create_time = create_time
        super().__init__(id=id, load_detail=load_detail)

    @classmethod
    def get_list(cls) -> list:
        """
        获取服务列表
        
        Returns
        -------
        List [Service]
            服务对象的集合。
        """
        res = get_with_token(f"{url_base()}/service/list")
        if res is None or not isinstance(res, list):
            raise AnyLearnException("请求未能得到有效响应")
        return [
            Service(id=s['id'], name=s['name'], description=s['description'],
                    state=s['state'], visibility=s['visibility'],
                    owner=s['owner'], model_id=s['model_id'],
                    creator_id=s['creator_id'], address=s['address'],
                    secret_key=s['secret_key'], create_time=s['create_time'],
                    gpu_num=s['gpu_num'], gpu_mem=s['gpu_mem'])
            for s in res
        ]

    def get_detail(self):
        """
        获取服务详细信息

        - 对象属性 :obj:`id` 应为非空

        Returns
        -------
        Service
            服务对象。
        """
        self._check_fields(required=['id'])
        res = get_with_token(f"{url_base()}/service/query",
                             params={'id': self.id})
        if not res or not isinstance(res, list):
            raise AnyLearnException("请求未能得到有效响应")
        res = res[0]
        self.__init__(id=res['id'], name=res['name'],
                      description=res['description'],
                      visibility=res['visibility'], model_id=res['model_id'],
                      address=res['address'], secret_key=res['secret_key'],
                      creator_id=res['creator_id'], envs=res['envs'],
                      replicas=res['replicas'], gpu_num=res['gpu_num'],
                      gpu_mem=res['gpu_mem'], state=res['state'],
                      create_time=res['create_time'], owner=res['owner'])

    def scale(self, replicas: int):
        """
        模型服务伸缩接口

        - 对象属性 :obj:`id` 应为非空

        Parameters
        ----------
        replicas : :obj:`int`
            模型服务副本数量(replicas>=1)。

        Returns
        -------
        bool
            True or False
        """
        if replicas < 1:
            raise AnyLearnException("服务副本数量不应小于1")
        self._check_fields(required=['id'])
        res = get_with_token(f"{url_base()}/service/scale",
                             params={
                                 'id': self.id,
                                 'replicas': replicas,
                             })
        if not res or 'data' not in res:
            raise AnyLearnException("请求未能得到有效响应")
        self.replicas = replicas
        return res.get('data', None) == self.id

    def stop(self):
        """
        模型服务停止接口

        - 对象属性 :obj:`id` 应为非空

        Returns
        -------
        bool
            True or False
        """
        self._check_fields(required=['id'])
        res = get_with_token(f"{url_base()}/service/stop",
                             params={
                                 'id': self.id,
                             })
        if not res or 'data' not in res:
            raise AnyLearnException("请求未能得到有效响应")
        self.state = ServiceState.STOPPED
        return res.get('data', None) == self.id

    def restart(self):
        """
        模型服务重启接口

        - 对象属性 :obj:`id` 应为非空

        Returns
        -------
        bool
            True or False
        """
        self._check_fields(required=['id'])
        res = get_with_token(f"{url_base()}/service/restart",
                             params={
                                 'id': self.id,
                             })
        if not res or 'data' not in res:
            raise AnyLearnException("请求未能得到有效响应")
        self.state = ServiceState.RUNNING
        return res.get('data', None) == self.id

    def get_deployment_status(self):
        """
        模型服务状态查询接口

        - 对象属性 :obj:`id` 应为非空

        :return: 
            .. code-block:: json

                {
                  "pod_names": [
                      "deployment-serv001",
                      "deployment-serv002",
                      "deployment-serv003"
                    ],
                  "workers": {
                    "available_workers": 1,
                    "replicas": 1,
                    "unavailable_workers": 0
                    }
                }
        """
        self._check_fields(required=['id'])
        res = get_with_token(f"{url_base()}/service/status",
                             params={
                                 'id': self.id,
                             })
        if not res or 'pods' not in res or 'workers' not in res:
            raise AnyLearnException("请求未能得到有效响应")
        res = {
            'pod_names': [p['name'] for p in res['pods']],
            'workers': res['workers'],
        }
        return res

    def get_log(self, pod_name, limit=100, direction="init", offset=0, includes=False):
        """
        模型服务日志查询接口

        - 对象属性 :obj:`id` 应为非空

        :param pod_name: :obj:`str`
                    容器单位名称。
        :param limit: :obj:`int`
                    日志条数上限（默认值100）。
        :param direction: :obj:`str`
                    日志查询方向。
        :param offset: :obj:`int`
                    日志查询索引。
        :param includes: :obj:`bool`
                    是否包含指定索引记录本身。

        :return: 
            .. code-block:: json

                [
                  {
                    "offset": 6554,
                    "text": "[Xlearn-Serving SERV123 ] - 2020-11-26 01:55:23,660 - DEBUG - (worker:1) is waiting"
                  }
                ]
        """
        self._check_fields(required=['id'])
        res = get_with_token(f"{url_base()}/service/log",
                             params={
                                 'id': self.id,
                                 'pod_name': pod_name,
                                 'limit': limit,
                                 'direction': direction,
                                 'index': offset,
                                 'self': 1 if includes else 0,
                             })
        if not res:
            raise AnyLearnException("请求未能得到有效响应")
        return res

    def get_records(self, page=1, size=20, load_detail=False):
        """
        模型服务运行结果查询接口

        - 对象属性 :obj:`id` 应为非空

        Parameters
        ----------
        page: :obj:`sintr`
            页面索引（默认值1）。
        size: :obj:`int`
            每页结果数量（默认值20）。
        load_detail: :obj:`bool`
            是否加载记录详情（默认为False）。

        Returns
        -------
        List [ServiceRecord]
            服务运行记录结果的集合。
        """
        self._check_fields(required=['id'])
        ress = get_with_token(f"{url_base()}/service/record",
                              params={
                                  'id': self.id,
                                  'page_index': page,
                                  'page_size': size,
                              })
        if not ress or not isinstance(ress, list):
            raise AnyLearnException("请求未能得到有效响应")
        return [
            ServiceRecord(id=res['id'], service_id=res['service_id'],
                          inference_data_file_id=res['file_id'],
                          error=res['error'], state=res['state'],
                          create_time=res['create_time'],
                          finish_time=res['finish_time'],
                          load_detail=load_detail)
            for res in ress
        ]

    def predict_online(self, file_binary: str):
        """
        在线预测接口

        Parameters
        ----------
        file_binary: :obj:`str`
            预测的文件。

        Returns
        -------
        ServiceRecord.result
            结果示例: :obj:`[[0.06298828125, 0.11171875149011612, 0.44580078125, 0.35546875, 5, 0.9521484375]]`
        ServiceRecord
            服务记录。
        """
        # Ensure service's address
        if not self.address:
            self.get_detail()
        
        # Call service's predict_online API
        url = lambda :AnylearnConfig.cluster_address + \
                self.address + \
                "/predict_online"
        res = requests.post(url(), files={'file': file_binary})
        res.raise_for_status()
        res.encoding = "utf-8"
        res = res.json()

        if not res or 'data' not in res:
            raise AnyLearnException("请求未能得到有效响应")
        
        # Fetch prediction results
        record = ServiceRecord(id=res['data'])
        while record.state not in [ServiceRecordState.FINISHED,
                                   ServiceRecordState.FAILED]:
            record.get_detail()
        filename = record.inference_data_file_id
        if not record.result:
            return "", record
        results = json.loads(json.loads(record.result)[filename])['object']
        return results, record

    def _namespace(self):
        return "service"



from datetime import datetime
import json
import time
from typing import Dict, List, Optional, Union

from anylearn.utils.api import (
    url_base,
    get_with_token,
    post_with_token,
    post_with_secret_key,
)
from anylearn.utils.errors import (
    AnyLearnException,
    AnyLearnMissingParamException,
)
from anylearn.utils.func import logs_beautify
from anylearn.interfaces.base import BaseObject
from anylearn.interfaces.resource import (
    AsyncResourceDownloader,
    File,
    Model,
    Resource,
    ResourceDownloader,
    ResourceState,
    ResourceVisibility,
    SyncResourceDownloader,
)
from anylearn.applications.train_profile import TrainProfile
from anylearn.storage.db.models import SqlLocalTrainTask


class TrainTaskState:
    """
    训练任务状态标识：

    - 0(CREATED)表示已创建
    - 1(RUNNING)表示运行中
    - 2(SUCCESS)表示已完成
    - -1(DELETED)表示已删除
    - -2(FAIL)表示失败
    - -3(ABORT)表示中断
    """
    CREATED = 0
    RUNNING = 1
    SUCCESS = 2
    DELETED = -1
    FAIL = -2
    ABORT = -3


class TrainTask(BaseObject):
    """
    AnyLearn训练任务类，以方法映射训练任务CRUD相关接口

    Attributes
    ----------
    id
        训练任务的唯一标识符，自动生成，由TRAI+uuid1生成的编码中后28个有效位（小写字母和数字）组成
    name
        训练任务的名称(长度1~50）)
    description
        训练任务描述（长度最大255）
    state
        训练任务状态
    visibility
        训练任务可见性，和所在训练项目可见性一致
    creator_id
        创建者ID
    owner
        训练任务的所有者，和所在训练项目所有者一致
    project_id
        训练项目ID
    algorithm_id
        算法ID
    algorithm_git_ref
        算法Gitea代码仓库的版本号（可以是commit号、分支名、tag名），非必填。
    train_params
        训练任务参数列表
    files
        训练相关的文件（默认'' 多个的话用逗号','隔开）
    results_id
        训练结果文件ID
    secret_key
        密钥
    create_time
        创建时间
    finish_time
        结束时间
    envs
        训练任务环境变量（默认''）
    hpo
        是否开启超参数自动调优
    hpo_search_space
        开启超参数自动调优时不能为空
    final_metric
        最终结果指标
    load_detail
        初始化时是否加载详情
    resource_request : :obj:`List[Dict[str, Dict[str, int]]]`, optional
        训练所需计算资源的请求。
        如未填，则使用Anylearn后端的:obj:`default`资源组中的默认资源套餐。
    """

    _fields = {
        # 资源创建/更新请求包体中必须包含且不能为空的字段
        'required': {
            'create': ['name', 'project_id', 'algorithm_id', 'train_params'],
            'update': [],
        },
        # 资源创建/更新请求包体中包含的所有字段
        'payload': {
            'create': ['name', 'description', 'project_id', 'algorithm_id',
                       'algorithm_git_ref', 'train_params', 'envs',
                       'files', 'visibility', 'resource_request'],
            'update': [],
        },
    }
    """
    创建/更新对象时：

    - 必须包含且不能为空的字段 :obj:`_fields['required']`
    - 所有字段 :obj:`_fields['payload']`
    """

    def __init__(self,
                 id: Optional[str]=None,
                 name: Optional[str]=None,
                 description: Optional[str]=None,
                 state: Optional[int]=None,
                 visibility: Optional[int]=None,
                 creator_id: Optional[str]=None,
                 owner: Optional[list]=None,
                 project_id: Optional[str]=None,
                 algorithm_id: Optional[str]=None,
                 algorithm_git_ref: Optional[str]=None,
                 train_params: Optional[str]=None,
                 files: Optional[list]=None,
                 results_id: Optional[str]=None,
                 secret_key: Optional[str]=None,
                 create_time: Optional[datetime]=None,
                 finish_time: Optional[datetime]=None,
                 envs: Optional[str]=None,
                 hpo=False,
                 hpo_search_space: Optional[str]=None,
                 final_metric: Optional[float]=None,
                 resource_request: Optional[List[Dict[str, Dict[str, int]]]]=None,
                 load_detail=False):
        """
        Parameters
        ----------
        id
            训练任务的唯一标识符，自动生成，由TRAI+uuid1生成的编码中后28个有效位（小写字母和数字）组成
        name
            训练任务的名称(长度1~50）)
        description
            训练任务描述（长度最大255）
        state
            训练任务状态
        visibility
            训练任务可见性，和所在训练项目可见性一致
        creator_id
            创建者ID
        owner
            训练任务的所有者，和所在训练项目所有者一致
        project_id
            训练项目ID
        algorithm_id
            算法ID
        algorithm_git_ref
            算法Gitea代码仓库的版本号（可以是commit号、分支名、tag名），非必填。
        train_params
            训练任务参数列表
        files
            训练相关的文件（默认'' 多个的话用逗号','隔开）
        results_id
            训练结果文件ID
        secret_key
            密钥
        create_time
            创建时间
        finish_time
            结束时间
        envs
            训练任务环境变量（默认''）
        hpo
            是否开启超参数自动调优
        hpo_search_space
            开启超参数自动调优时不能为空
        final_metric
            最终结果指标
        resource_request : :obj:`List[Dict[str, Dict[str, int]]]`, optional
            训练所需计算资源的请求。
            如未填，则使用Anylearn后端的:obj:`default`资源组中的默认资源套餐。
        load_detail
            初始化时是否加载详情
        """
        self.name = name
        self.description = description
        self.state = state
        self.visibility = visibility
        self.creator_id = creator_id
        self.owner = owner
        self.project_id = project_id
        self.algorithm_id = algorithm_id
        self.algorithm_git_ref = algorithm_git_ref
        self.train_params = train_params
        self.files = files
        self.results_id = results_id
        self.secret_key = secret_key
        self.create_time = create_time
        self.finish_time = finish_time
        self.envs = envs
        self.hpo = hpo
        self.hpo_search_space = hpo_search_space
        self.final_metric = final_metric
        self.resource_request = resource_request
        super().__init__(id=id, load_detail=load_detail)

    def finished(self):
        """
        检查训练任务是否完成

        Returns
        -------
        bool
            True or False
        """
        return self.state in [
            TrainTaskState.SUCCESS,
            TrainTaskState.FAIL,
            TrainTaskState.ABORT,
            TrainTaskState.DELETED,
        ]

    @classmethod
    def get_list(cls):
        """
        Listing is not supported for TrainTask
        """
        raise AnyLearnException("Listing is not supported for TrainTask")

    def get_detail(self):
        """
        获取训练任务详细信息

        - 对象属性 :obj:`id` 应为非空

        Returns
        -------
        TrainTask
            训练任务对象。
        """
        self._check_fields(required=['id'])
        res = get_with_token(f"{url_base()}/train_task/query",
                             params={'id': self.id})
        if not res or not isinstance(res, list):
            raise AnyLearnException("请求未能得到有效响应")
        res = res[0]
        self.__init__(id=res['id'], name=res['name'],
                      description=res['description'], state=res['state'],
                      visibility=res['visibility'],
                      creator_id=res['creator_id'], owner=res['owner'],
                      project_id=res['project_id'],
                      algorithm_id=res['algorithm_id'],
                      algorithm_git_ref=res['algorithm_git_ref'],
                      train_params=res['args'], files=res['files'],
                      results_id=res['results_id'],
                      secret_key=res['secret_key'],
                      create_time=res['create_time'],
                      finish_time=res['finish_time'],
                      envs=res['envs'],
                      resource_request=json.loads(res['resource_request']))

    def _create(self):
        data = self._payload_create()
        if data['files'] and isinstance(data['files'], list):
            data['files'] = ','.join(data['files'])
        if self.hpo:
            if not self.hpo_search_space:
                msg = f"{self.__class__.__name__}缺少必要字段：hpo_search_space" + \
                    "——当开启超参数自动调优时（hpo==True），hpo_search_space为必填字段"
                raise AnyLearnMissingParamException(msg)
            data['hpo'] = 1
            data['hpo_search_space'] = json.dumps(self.hpo_search_space)
        if data['resource_request']:
            data['resources'] = json.dumps(data['resource_request'])
        del(data['resource_request'])
        res = post_with_token(self._url_create(), data=data)
        if not res or 'data' not in res:
            raise AnyLearnException("请求未能得到有效响应")
        self.id = res['data']
        return True

    def _update(self):
        # No update for train task
        pass

    def get_log(self, limit=100, direction="init", offset=0, offset_index=-1):
        """
        训练任务日志查询接口

        - 对象属性 :obj:`id` 应为非空

        :param limit: :obj:`int`
            日志条数上限（默认值100）。
        :param direction: :obj:`str`
            日志查询方向。
        :param offset: :obj:`int`
            日志查询偏移量。
        :param offset_index :obj:`int`
            日志查询偏移量索引，搭配偏移量使用作为分页基准。

        :return: 

            .. code-block:: json

                [
                  {
                    "offset": 164324567,
                    "offset_index": 1234,
                    "text": "Task TRAId123 started."
                  },
                  {
                    "offset": 164324590,
                    "offset_index": 1238,
                    "text": "Task TRAId123 finished."
                  }
                ]

        """
        self._check_fields(required=['id'])
        params = {
            'id': self.id,
            'limit': limit,
            'direction': direction,
            'index': offset,
            'offset_index': offset_index,
        }
        res = get_with_token(f"{url_base()}/train_task/log", params=params)
        if not res or type(res) != list:
            raise AnyLearnException("请求未能得到有效响应")
        return [r for r in res if r['text'].strip()]

    def get_last_log(self, limit: int=100, debug: bool=False):
        """
        训练任务日志最近n行查询接口，返回日志文本列表。

        - 对象属性 :obj:`id` 应为非空

        :param limit :obj:`int`
            需要查询的行数（默认100）。
        :param debug :obj:`bool`
            是否显示更全面的debug信息（默认False）。

        :return: 

            .. code-block:: json

                [
                  "log text1",
                  "log text2"
                ]

        """
        logs = logs_beautify(logs=self.get_log(limit=limit), debug=debug)
        return list(reversed(logs))

    def get_full_log(self, debug: bool=False):
        """
        训练任务日志全量查询接口，返回日志文本列表。

        - 对象属性 :obj:`id` 应为非空

        :param debug :obj:`bool`
            是否显示更全面的debug信息（默认False）。

        :return: 

            .. code-block:: json

                [
                  "log text1",
                  "log text2"
                ]

        """
        logs = []
        offset = 0
        offset_index = -1
        while True:
            try:
                log_parts = self.get_log(
                    offset=offset,
                    offset_index=offset_index,
                    direction="back"
                )
                last = log_parts[-1]
                offset = last['offset']
                offset_index = last['offset_index']
                logs.extend(log_parts)
            except:
                break
        return logs_beautify(logs=logs, debug=debug)

    def stream_log(self,
                   init_limit: int=100,
                   polling: int=2,
                   debug: bool=False):
        """
        实时训练任务日志流式生成接口，每次迭代返回日志文本的一行。

        - 对象属性 :obj:`id` 应为非空

        :param init_limit :obj:`bool`
            起始日志需要查询的行数（默认100）。
        :param polling :obj:`int`
            轮询间隔时间（单位：秒，默认2）。
        :param debug :obj:`bool`
            是否显示更全面的debug信息（默认False）。

        :return: :obj:`Iterator`

        """
        # Some initial logs (latest)
        logs = []
        while not logs:
            logs = self.get_log(limit=init_limit)
        for l in reversed(logs_beautify(logs, debug=debug)):
            yield l
        # Keep fetching new logs
        offset = logs[0]['offset']
        offset_index = logs[0]['offset_index']
        while not self.finished():
            try:
                self.get_detail()
                logs = self.get_log(
                    offset=offset,
                    offset_index=offset_index,
                    direction="back",
                )
                if not logs:
                    continue
                for l in logs_beautify(logs, debug=debug):
                    yield l
                offset = logs[-1]['offset']
                offset_index = logs[-1]['offset_index']
            except:
                time.sleep(polling)

    def get_status(self):
        """
        训练任务状态查询接口

        - 对象属性 :obj:`id` 、 :obj:`secret_key` 应为非空

        :return: 
            .. code-block:: json

                {
                  "current_epoch": "2",
                  "current_train_loss": "2.169192314147949",
                  "current_train_step": "1288",
                  "ip": "10.244.2.124",
                  "process": "1.0",
                  "secret_key": "TKEY123",
                  "state": "success"
                }

        """
        self._check_fields(required=['id', 'secret_key'])
        params = {
            'id': self.id,
            'secret_key': self.secret_key,
        }
        res = get_with_token(f"{url_base()}/train_task/status", params=params)
        if not res or type(res) != dict:
            raise AnyLearnException("请求未能得到有效响应")
        return res

    def download_results(self,
                         save_path: str,
                         async_download: bool=True,
                         downloader: Optional[ResourceDownloader]=None,
                         polling: Union[float, int]=5,
                        ):
        """
        下载训练任务结果

        Parameters
        ----------
        save_path : :obj:`str`
            文件保存路径。
        downloader : :obj:`ResourceDownloader`
            可以使用SDK中的AsyncResourceDownloader，也可以自定义实现ResourceDownloader。
        polling : :obj:`float, int`
            下载前要先压缩文件，轮询查看文件有没有压缩完的时间间隔，单位：秒。默认值5

        Returns
        -------
        str
            文件名。
        """
        self._check_fields(required=['results_id'])
        if self.state == TrainTaskState.FAIL:
            raise AnyLearnException("训练失败！")
        elif self.state == TrainTaskState.DELETED:
            raise AnyLearnException("训练已删除！")
        elif self.state == TrainTaskState.ABORT:
            raise AnyLearnException("训练中断！")
        elif self.state == TrainTaskState.CREATED:
            raise AnyLearnException("训练未开始!")
        elif self.state == TrainTaskState.RUNNING:
            raise AnyLearnException("正在训练中，请耐心等待...")
        if not downloader:
            if async_download:
                downloader = AsyncResourceDownloader()
            else:
                downloader = SyncResourceDownloader()
        return Resource.download_file(resource_id=self.results_id, # type: ignore
                                      save_path=save_path,
                                      downloader=downloader,
                                      polling=polling,
                                     )

    def report_final_metric(self, metric: float):
        """
        训练任务汇报最终结果指标

        - 对象属性 :obj:`id` 、 :obj:`secret_key` 应为非空

        :param metric: :obj:`float`
            最终结果指标。

        :return:
            .. code-block:: json

                {
                  "msg": "任务TRAId123结果指标保存成功"
                }

        """
        self._check_fields(required=['id', 'secret_key'])
        data = {
            'id': self.id,
            'metric': metric,
        }
        res = post_with_secret_key(f"{url_base()}/train_task/final_metric",
                                   data=data,
                                   secret_key=self.secret_key)
        if not res or type(res) != dict:
            raise AnyLearnException("请求未能得到有效响应")
        self.final_metric = metric
        return res

    def get_final_metric(self):
        """
        获取训练任务最终结果指标

        - 对象属性 :obj:`id` 应为非空

        :return: 
            .. code-block:: json

                {
                  "final_metric": 662.8,
                  "id": "TRAI1d3",
                  "name": "test"
                }

        """
        self._check_fields(required=['id'])
        res = get_with_token(f"{url_base()}/train_task/final_metric",
                             params={'id': self.id})
        if not res or 'final_metric' not in res:
            raise AnyLearnException("请求未能得到有效响应")
        self.final_metric = res['final_metric']
        return res

    def report_intermediate_metric(self, metric: float):
        """
        训练任务汇报中间结果指标

        - 对象属性 :obj:`id` 、 :obj:`secret_key` 应为非空

        :param metric: :obj:`float`
            中间结果指标。

        :return: 
            .. code-block:: json

                {
                  "msg": "任务TRAId123结果指标保存成功"
                }

        """
        self._check_fields(required=['id', 'secret_key'])
        data = {
            'id': self.id,
            'metric': metric,
        }
        res = post_with_secret_key(f"{url_base()}/train_task/intermediate_metric",
                                   data=data,
                                   secret_key=self.secret_key)
        if not res or type(res) != dict:
            raise AnyLearnException("请求未能得到有效响应")
        return res

    def get_intermediate_metric(self, last_timestamp: str="1970-01-01 00:00:00"):
        """
        获取训练任务中间结果指标

        - 对象属性 :obj:`id` 应为非空

        :param last_timestamp: :obj:`str`
            仅获取某一时刻以后的中间结果,格式"1970-01-01 00:00:00"

        :return: 
            .. code-block:: json

                [
                  {
                    "id": "METR123",
                    "metric": 90.0,
                    "train_task_id": "TRAI123",
                    "reported_at": "2021-04-29 21:00:00"
                  },
                  {
                    "id": "METR456",
                    "metric": 99.0,
                    "train_task_id": "TRAI123",
                    "reported_at": "2021-04-29 21:10:00"
                  }
                ]

        """
        self._check_fields(required=['id'])
        res = get_with_token(f"{url_base()}/train_task/intermediate_metric",
                             params={
                                 'id': self.id,
                                 'last_timestamp': last_timestamp,
                             })
        if not res or type(res) != list:
            raise AnyLearnException("请求未能得到有效响应")
        return res

    def transform_model(self,
                        file_path: str,
                        name: str,
                        description: Optional[str]=None,
                        is_zipfile: bool=False,
                        visibility: int=ResourceVisibility.PRIVATE,
                        owner: Optional[list]=None,
                        polling: Union[float, int]=5):
        """
        模型文件转存接口

        - 对象属性 :obj:`results_id` 、 :obj:`algorithm_id` 应为非空

        Parameters
        ----------
        file_path : :obj:`str`
            模型文件路径。
        name : :obj:`str`
            模型的名称。
        description : :obj:`str`
            模型描述。
        is_zipfile : :obj:`bool`
            模型是否为zip文件。默认为False。
        visibility : :obj:`int`
            模型的可见性,默认为私有PRIVATE。
        owner : :obj:`list`
            模型的所有者。
        polling : :obj:`float|int`
            模型转换中轮询模型状态的时间间隔（单位：秒）。 默认为5秒。

        Returns
        -------
        Model
            模型对象。
        """
        self._check_fields(required=['results_id', 'algorithm_id'])
        data = {
            'file_id': self.results_id,
            'file_path': file_path,
            'name': name,
            'description': description,
            'algorithm_id': self.algorithm_id,
            'is_zipfile': "1" if is_zipfile else "0",
            'visibility': visibility,
        }
        if isinstance(owner, list) and len(owner):
            data['owner'] = ','.join(owner)
        res = post_with_token(f"{url_base()}/model/transform",
                              data=data)
        if not res or 'data' not in res:
            raise AnyLearnException("请求未能得到有效响应")
        model = Model(id=res['data'])
        finished = [ResourceState.ERROR, ResourceState.READY]
        while model.state not in finished:
            time.sleep(polling)
            model.get_detail()
        if model.state == ResourceState.ERROR:
            raise AnyLearnException("Error occured when transforming model")
        return model

    def get_train_profile(self):
        """
        获取训练任务描述

        - 对象属性 :obj:`id` 应为非空

        Returns
        -------
        TrainProfile
            训练任务描述对象。
        """
        self._check_fields(required=['id'])
        return TrainProfile.get(train_task_id=self.id)

    def get_results_file(self) -> File:
        self._check_fields(required=['results_id'])
        return File(id=self.results_id, load_detail=True)

    @classmethod
    def from_sql(cls, sql_local_train_task: SqlLocalTrainTask):
        """
        把本地保存的训练任务SqlLocalTrainTask转化为TrainTask

        Parameters
        ----------
        sql_local_train_task : :obj:`SqlLocalTrainTask`
            本地训练任务

        Returns
        -------
        TrainTask
        """
        return TrainTask(id=sql_local_train_task.id, state=int(sql_local_train_task.remote_state_sofar),
                         secret_key=sql_local_train_task.secret_key,
                         project_id=sql_local_train_task.project_id,
                         final_metric=sql_local_train_task.hpo_final_metric) # type: ignore

    def _namespace(self):
        return "train_task"



"""Alter resource tables for checksum

Revision ID: 848203d98d73
Revises: 
Create Date: 2021-05-13 15:40:20.995359

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.sql.schema import Column


# revision identifiers, used by Alembic.
revision = '848203d98d73'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'local_algorithm',
        sa.Column('id', sa.String(length=32), primary_key=True),
        sa.Column('checksum', sa.String(length=128), nullable=False),
    )
    op.create_table(
        'local_dataset',
        sa.Column('id', sa.String(length=32), primary_key=True),
        sa.Column('checksum', sa.String(length=128), nullable=False),
    )
    op.create_table(
        'local_model',
        sa.Column('id', sa.String(length=32), primary_key=True),
        sa.Column('checksum', sa.String(length=128), nullable=False),
    )
    op.create_table(
        'local_train_task',
        sa.Column('id', sa.String(length=32), primary_key=True),
        sa.Column('remote_state_sofar', sa.Integer, nullable=False),
        sa.Column('secret_key', sa.String(length=32), nullable=False),
        sa.Column('project_id', sa.String(length=32), nullable=False),
    )


def downgrade():
    # WARNING: all data will be lost
    op.drop_table('local_algorithm')
    op.drop_table('local_dataset')
    op.drop_table('local_model')
    op.drop_table('local_train_task')



import responses
from requests.exceptions import RequestException

from anylearn.config import init_sdk
from anylearn.interfaces.resource.resource import ResourceState, ResourceVisibility
from anylearn.interfaces.resource.file import File
from anylearn.utils.errors import AnyLearnException, AnyLearnMissingParamException
from tests.base_test_case import BaseTestCase

class TestFile(BaseTestCase):
    @responses.activate
    def test_list_file_ok(self):
        responses.add(responses.GET, url=self._url("file/list"),
                      json=[
                          {
                              'id': "FILE001",
                              'name': "TestFile1",
                              'description': "test",
                              'state': ResourceState.READY,
                              'visibility': ResourceVisibility.PUBLIC,
                              'upload_time': "2020-01-01 00:00",
                              'filename': "testfile1.txt",
                              'creator_id': "USER001",
                              'node_id': "NODE001",
                              'owner': ["USER001"],
                          },
                          {
                              'id': "FILE002",
                              'name': "TestFile2",
                              'description': "test",
                              'state': ResourceState.READY,
                              'visibility': ResourceVisibility.PUBLIC,
                              'upload_time': "2020-01-02 00:00",
                              'filename': "testfile2.txt",
                              'creator_id': "USER002",
                              'node_id': "NODE002",
                              'owner': ["USER002"],
                          },
                      ],
                      status=200)
        files = File.get_list()
        self.assertIsInstance(files, list)
        self.assertEqual(len(files), 2)
        self.assertIsInstance(files[0], File)
        self.assertIsInstance(files[1], File)
        self.assertEqual(files[0].id, "FILE001")
        self.assertEqual(files[1].id, "FILE002")
        self.assertEqual(files[0].filename, "testfile1.txt")
        self.assertEqual(files[1].filename, "testfile2.txt")

    @responses.activate
    def test_list_file_ko_unknown_response(self):
        responses.add(responses.GET, url=self._url("file/list"),
                      json={'msg': "Unknown response"},
                      status=200)
        with self.assertRaises(AnyLearnException) as ctx:
            File.get_list()
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnException)
        self.assertEqual(e.msg, "请求未能得到有效响应")

    @responses.activate
    def test_get_file_detail_explicit_ok(self):
        responses.add(responses.GET, url=self._url("file/query?id=FILE001"),
                      match_querystring=True, json=[{
                          'id': "FILE001",
                          'name': "TestFile1",
                          'description': "test",
                          'state': ResourceState.READY,
                          'visibility': ResourceVisibility.PUBLIC,
                          'upload_time': "2020-01-01 00:00",
                          'filename': "test.tar.gz",
                          'is_zipfile': 1,
                          'file': "USER001/files/FILE001/test.tar.gz",
                          'size': "250.41",
                          'creator_id': "USER001",
                          'node_id': "NODE001",
                          'owner': ["USER001"],
                      }],
                      status=200)
        test_file = File(id="FILE001")
        test_file.get_detail()
        self.assertEqual(test_file.filename, "test.tar.gz")
        self.assertTrue(test_file.is_zipfile)
        self.assertEqual(test_file.file_path, "USER001/files/FILE001/test.tar.gz")
        self.assertEqual(test_file.size, "250.41")
        self.assertEqual(test_file.creator_id, "USER001")
        self.assertEqual(test_file.node_id, "NODE001")
        self.assertIsInstance(test_file.owner, list)
        self.assertEqual(test_file.owner[0], "USER001")

    @responses.activate
    def test_get_file_detail_no_explicit_ok(self):
        responses.add(responses.GET, url=self._url("file/query?id=FILE001"),
                      match_querystring=True, json=[{
                          'id': "FILE001",
                          'name': "TestFile1",
                          'description': "test",
                          'state': ResourceState.READY,
                          'visibility': ResourceVisibility.PUBLIC,
                          'upload_time': "2020-01-01 00:00",
                          'filename': "test.tar.gz",
                          'is_zipfile': 1,
                          'file': "USER001/files/FILE001/test.tar.gz",
                          'size': "250.41",
                          'creator_id': "USER001",
                          'node_id': "NODE001",
                          'owner': ["USER001"],
                      }],
                      status=200)
        test_file = File(id="FILE001", load_detail=True)
        self.assertEqual(test_file.name, "TestFile1")
        self.assertEqual(test_file.filename, "test.tar.gz")
        self.assertTrue(test_file.is_zipfile)
        self.assertEqual(test_file.file_path, "USER001/files/FILE001/test.tar.gz")
        self.assertEqual(test_file.size, "250.41")
        self.assertEqual(test_file.creator_id, "USER001")
        self.assertEqual(test_file.node_id, "NODE001")
        self.assertIsInstance(test_file.owner, list)
        self.assertEqual(test_file.owner[0], "USER001")

    @responses.activate
    def test_get_file_detail_ko_403(self):
        responses.add(responses.GET, url=self._url("file/query?id=FILE403"),
                      match_querystring=True, status=403)
        test_file = File(id="FILE403")
        with self.assertRaises(RequestException) as ctx:
            test_file.get_detail()
        e = ctx.exception
        self.assertIsInstance(e, RequestException)
        self.assertEqual(e.response.status_code, 403)

    @responses.activate
    def test_get_file_detail_ko_unknown_response(self):
        responses.add(responses.GET, url=self._url("file/query?id=FILE250"),
                      match_querystring=True, json={'msg': "Unknown response"},
                      status=200)
        test_file = File(id="FILE250")
        with self.assertRaises(AnyLearnException) as ctx:
            test_file.get_detail()
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnException)
        self.assertEqual(e.msg, "请求未能得到有效响应")

    @responses.activate
    def test_get_file_detail_empty_list(self):
        responses.add(responses.GET, url=self._url("file/query?id=FILE250"),
                      match_querystring=True, json=[],
                      status=200)
        test_file = File(id="FILE250")
        with self.assertRaises(AnyLearnException) as ctx:
            test_file.get_detail()
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnException)
        self.assertEqual(e.msg, "请求未能得到有效响应")

    @responses.activate
    def test_create_file_ok(self):
        responses.add(responses.POST, url=self._url("file/add"),
                      json={'data': "FILE001", 'message': "文件添加成功"},
                      status=200)
        test_file = File(name="TestFILE001", filename="test.tar.gz")
        result = test_file.save()
        self.assertTrue(result)
        self.assertEqual(test_file.id, "FILE001")

    @responses.activate
    def test_create_file_ko_empty_name_filename(self):
        test_file = File()
        with self.assertRaises(AnyLearnMissingParamException) as ctx:
            test_file.save()
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnMissingParamException)
        self.assertEqual(e.msg, "File缺少必要字段：['name', 'filename']")

    @responses.activate
    def test_create_file_ko_unknown_response(self):
        responses.add(responses.POST, url=self._url("file/add"),
                      json={'msg': "Unknown response"}, status=201)
        test_file = File(name="TestFile", filename="FILE250.tar.gz")
        with self.assertRaises(AnyLearnException) as ctx:
            test_file.save()
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnException)
        self.assertEqual(e.msg, "请求未能得到有效响应")

    @responses.activate
    def test_update_file_ok(self):
        responses.add(responses.PUT, url=self._url("file/update"),
                      json={'data': "FILE001", 'message': "文件信息更新成功"},
                      status=200)
        test_file = File(id="FILE001", name="TestFile1", description="test m",
                         state=ResourceState.READY,
                         visibility=ResourceVisibility.PUBLIC,
                         upload_time="2020-01-01 00:00")
        result = test_file.save()
        self.assertTrue(result)

    @responses.activate
    def test_update_file_ko_empty_name(self):
        test_file = File(id="FILE250", name="")
        with self.assertRaises(AnyLearnMissingParamException) as ctx:
            test_file.save()
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnMissingParamException)
        self.assertEqual(e.msg, "File缺少必要字段：['name']")

    @responses.activate
    def test_update_file_ko_unknown_response(self):
        responses.add(responses.PUT, url=self._url("file/update"),
                      json={'msg': "Unknown response"},
                      status=200)
        test_file = File(id="FILE250", name="TestFile", description="test",
                         state=ResourceState.READY,
                         visibility=ResourceVisibility.PROTECTED,
                         upload_time="2020-01-01 02:50")
        with self.assertRaises(AnyLearnException) as ctx:
            test_file.save()
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnException)
        self.assertEqual(e.msg, "请求未能得到有效响应")

    @responses.activate
    def test_delete_file_ok(self):
        responses.add(responses.DELETE,
                      url=self._url("file/delete?id=FILE001&force=0"),
                      match_querystring=True,
                      json={'data': "FILE001", 'message': "文件删除成功"},
                      status=200)
        test_file = File(id="FILE001", name="TestFile1", description="test d",
                         state=ResourceState.READY,
                         visibility=ResourceVisibility.PUBLIC,
                         upload_time="2020-01-01 00:00")
        result = test_file.delete()
        self.assertTrue(result)

    @responses.activate
    def test_delete_file_ko_unknown_response(self):
        responses.add(responses.DELETE,
                      url=self._url("file/delete?id=FILE250&force=0"),
                      match_querystring=True,
                      json={'msg': "Unknown response"},
                      status=204)
        test_file = File(id="FILE250", name="TestFile", description="test",
                         state=ResourceState.READY,
                         visibility=ResourceVisibility.PROTECTED,
                         upload_time="2020-01-01 02:50")
        with self.assertRaises(AnyLearnException) as ctx:
            test_file.delete()
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnException)
        self.assertEqual(e.msg, "请求未能得到有效响应")



from datetime import datetime
from typing import Optional, Union

from anylearn.utils.api import url_base, get_with_token
from anylearn.utils.errors import AnyLearnException
from anylearn.interfaces.resource.resource import Resource

class File(Resource):
    """
    AnyLearn其他文件类，以方法映射文件CRUD相关接口
    
    Attributes
    ----------
    id
        文件的唯一标识符，自动生成，由FILE+uuid1生成的编码中后28个有效位（小写字母和数字）组成（自动生成）
    name
        文件的名称（长度1~50）
    description
        文件描述（长度最大255）
    state
        文件状态
    visibility
        文件的可见性（默认为3）
    upload_time
        文件上传时间
    filename
        下一步中会被分片上传的文件的完整文件名（包括扩展名）（长度1~128）
    is_zipfile
        是否为zip文件
    file_path
        文件文件路径
    size
        文件大小
    creator_id
        创建者ID
    node_id
        文件节点ID
    owner
        文件的所有者，以逗号分隔的这些用户的ID拼成的字符串，无多余空格
    load_detail
        初始化时是否加载详情
    """

    """具体资源信息配置"""
    _fields = {
        # 资源创建/更新请求包体中必须包含且不能为空的字段
        'required': {
            'create': ['name', 'filename'],
            'update': ['id', 'name'],
        },
        # 资源创建/更新请求包体中包含的所有字段
        'payload': {
            'create': ['name', 'description', 'visibility', 'owner',
                       'filename'],
            'update': ['id', 'name', 'description', 'visibility', 'owner'],
        },
    }
    """
    创建/更新对象时：

    - 必须包含且不能为空的字段 :obj:`_fields['required']`
    - 所有字段 :obj:`_fields['payload']`
    """

    def __init__(self,
                 id: Optional[str]=None,
                 name: Optional[str]=None,
                 description: Optional[str]=None,
                 state: Optional[int]=None,
                 visibility=3,
                 upload_time: Optional[Union[datetime, str]]=None,
                 filename: Optional[str]=None,
                 is_zipfile: Optional[int]=None,
                 file_path: Optional[str]=None,
                 size: Optional[str]=None,
                 creator_id: Optional[str]=None,
                 node_id: Optional[str]=None,
                 owner: Optional[list]=None,
                 load_detail=False):
        """
        Parameters
        ----------
        id
            文件的唯一标识符，自动生成，由FILE+uuid1生成的编码中后28个有效位（小写字母和数字）组成（自动生成）
        name
            文件的名称（长度1~50）
        description
            文件描述（长度最大255）
        state
            文件状态
        visibility
            文件的可见性（默认为3）
        upload_time
            文件上传时间
        filename
            下一步中会被分片上传的文件的完整文件名（包括扩展名）（长度1~128）
        is_zipfile
            是否为zip文件
        file_path
            文件文件路径
        size
            文件大小
        creator_id
            创建者ID
        node_id
            文件节点ID
        owner
            文件的所有者，以逗号分隔的这些用户的ID拼成的字符串，无多余空格
        load_detail
            初始化时是否加载详情
        """
        super().__init__(id=id, name=name, description=description,
                         state=state, visibility=visibility,
                         upload_time=upload_time, filename=filename,
                         is_zipfile=is_zipfile, file_path=file_path, size=size,
                         creator_id=creator_id, node_id=node_id, owner=owner,
                         load_detail=load_detail)

    @classmethod
    def get_list(cls) -> list:
        """
        获取文件列表
        
        Returns
        -------
        List [File]
            文件集合。
        """
        res = get_with_token(f"{url_base()}/file/list")
        if res is None or not isinstance(res, list):
            raise AnyLearnException("请求未能得到有效响应")
        return [
            File(id=f['id'], name=f['name'], description=f['description'],
                 state=f['state'], visibility=f['visibility'],
                 upload_time=f['upload_time'], filename=f['filename'],
                 creator_id=f['creator_id'], node_id=f['node_id'],
                 owner=f['owner'])
            for f in res
        ]

    def get_detail(self):
        """
        获取文件详细信息

        - 对象属性 :obj:`id` 应为非空

        Returns
        -------
        File
            文件对象。
        """
        self._check_fields(required=['id'])
        res = get_with_token(f"{url_base()}/file/query",
                             params={'id': self.id})
        if not res or not isinstance(res, list):
            raise AnyLearnException("请求未能得到有效响应")
        res = res[0]
        self.__init__(id=res['id'], name=res['name'],
                      description=res['description'], state=res['state'],
                      visibility=res['visibility'],
                      upload_time=res['upload_time'], filename=res['filename'],
                      is_zipfile=True if res['is_zipfile'] == 1 else False,
                      file_path=res['file'], size=res['size'],
                      creator_id=res['creator_id'], node_id=res['node_id'],
                      owner=res['owner'])

    def _namespace(self):
        return "file"



from typing import Optional

from anylearn.interfaces.base import BaseObject
from anylearn.utils.api import url_base, get_with_token
from anylearn.utils.errors import (
    AnyLearnException,
    AnyLearnMissingParamException,
)


class QuotaGroup(BaseObject):
    """
    AnyLearn资源组类

    Attributes
    ----------
    id
        资源组的唯一标识符，自动生成，由QGRP+uuid1生成的编码中后28个有效位（小写字母和数字）组成
    name
        资源组名称（非空 长度1~50）
    level
        资源等级（Guarantee或BestEffort）
    node_level
        节点匹配等级
    capacity
        资源组内各资源类型的容量
    default
        资源组默认的单次任务申请用量
    capacity
        资源组内各资源类型的实际用量
    load_detail
        初始化时是否加载详情
    """

    _fields = {
        'required': {
            'create': [],
            'update': [],
        },
        'payload': {
            'create': [],
            'update': [],
        },
    }
    """
    创建/更新对象时：

    - 必须包含且不能为空的字段 :obj:`_fields['required']`
    - 所有字段 :obj:`_fields['payload']`
    """

    def __init__(self,
                 id: Optional[str]=None,
                 name: Optional[str]=None,
                 level: Optional[str]=None,
                 node_level: Optional[str]=None,
                 capacity: Optional[dict]=None,
                 default: Optional[dict]=None,
                 allocated: Optional[dict]=None,
                 load_detail=False):
        """
        Parameters
        ----------
        id
            资源组的唯一标识符，自动生成，由QGRP+uuid1生成的编码中后28个有效位（小写字母和数字）组成
        name
            资源组名称（非空 长度1~50）
        level
            资源等级（Guarantee或BestEffort）
        node_level
            节点匹配等级
        load_detail
            初始化时是否加载详情
        capacity
            资源组内各资源类型的容量
        default
            资源组默认的单次任务申请用量
        capacity
            资源组内各资源类型的实际用量
        """
        self.id = id
        self.name = name
        self.level = level
        self.node_level = node_level
        self.capacity = capacity
        self.default = default
        self.allocated = allocated
        super().__init__(id=id, load_detail=load_detail)

    @classmethod
    def get_list(cls):
        """
        获取资源组列表
        
        Returns
        -------
        List [QuotaGroup]]
            资源组对象的集合。
        """
        res = get_with_token(f"{url_base()}/quota_group/list")
        if res is None or not isinstance(res, list):
            raise AnyLearnException("请求未能得到有效响应")
        return [
            QuotaGroup(
                id=item['id'],
                name=item['name'],
                level=item['level'],
                node_level=item['node_level'],
                capacity=item['capacity'],
                default=item['default'],
                allocated=item['allocated'],
            )
            for item in res
        ]

    def get_detail(self):
        """
        获取训练项目详细信息

        - 对象属性 :obj:`id` 应为非空

        Returns
        -------
        Project
            训练项目对象。
        """
        if not self.id and not self.name:
            raise AnyLearnMissingParamException("QuotaGroup缺少必要字段：id或name")
        params = {'id': self.id} if self.id else {'name': self.name}
        res = get_with_token(f"{url_base()}/quota_group/query", params=params)
        if not res or not isinstance(res, list):
            raise AnyLearnException("请求未能得到有效响应")
        res = res[0]
        self.__init__(
            id=res['id'],
            name=res['name'],
            level=res['level'],
            node_level=res['node_level'],
            capacity=res['capacity'],
            default=res['default'],
            allocated=res['allocated'],
        )

    def available(self):
        used = dict({k: 0 for k in self.capacity.keys()}, **self.allocated)
        return {k: self.capacity[k] - used[k] for k in self.capacity.keys()}

    def save(self):
        raise AnyLearnException("Saving for QuotaGroup is not supported in SDK")

    def delete(self):
        raise AnyLearnException("Deleting for QuotaGroup is not supported in SDK")

    def _namespace(self):
        return "quota_group"



from pathlib import Path
import responses
from requests.exceptions import RequestException
from unittest.mock import patch, create_autospec

import pytest

from anylearn.applications.quickstart import quick_evaluate
from anylearn.config import init_sdk
from anylearn.interfaces import EvaluateTask, EvaluateTaskState
from anylearn.interfaces.resource import (
    Algorithm,
    Dataset,
    Model,
    ResourceState,
    ResourceVisibility
)
from anylearn.utils.errors import (
    AnyLearnException,
    AnyLearnNotSupportedException
)
from tests.base_test_case import BaseTestCase


FAKE_ALGO_OK = Path(__file__).parent / "fake_algo_ok"
FAKE_ALGO_KO = Path(__file__).parent / "fake_algo_ko"
FAKE_DSET = Path(__file__).parent / "fake_dset"
FAKE_MODEL = Path(__file__).parent / "fake_model"


class TestQuickEvaluate(BaseTestCase):
    @pytest.mark.skip()
    @patch('anylearn.interfaces.resource.resource_uploader.ResourceUploader')
    @responses.activate
    def test_quick_evaluate_ok(self, MockResourceUploader):
        @create_autospec
        def mock_upload_ok(resource_id, chunks):
            return True

        # Mock uploader with stubbed upload function
        uploader = MockResourceUploader()
        uploader.run = mock_upload_ok

        # Stub mirror list
        responses.add(responses.GET, url=self._url("mirror/list"),
                      json=[
                          {
                              'id': "MIRR001",
                              'name': "Test001",
                              'requirements': "test==0.0.1",
                          },
                          {
                              'id': "MIRR002",
                              'name': "QUICKSTART",
                              'requirements': "test==0.0.2",
                          },
                      ],
                      status=200)
        # Stub algo creation
        responses.add(responses.POST, url=self._url("algorithm/add"),
                      json={'data': "ALGO001", 'message': "算法添加成功"},
                      status=200)
        # Stub model creation
        responses.add(responses.POST, url=self._url("model/add"),
                      json={'data': "MODE001", 'message': "模型添加成功"},
                      status=200)
        # Stub dset creation
        responses.add(responses.POST, url=self._url("dataset/add"),
                      json={'data': "DSET001", 'message': "数据集添加成功"},
                      status=200)
        # Stub algo upload finish
        responses.add(responses.POST, url=self._url("resource/upload_finish"),
                      json={
                          'msg': "上传已完成，稍后可查看上传结果"
                      },
                      status=200)
        # Stub model upload finish
        responses.add(responses.POST, url=self._url("resource/upload_finish"),
                      json={
                          'msg': "上传已完成，稍后可查看上传结果"
                      },
                      status=200)
        # Stub dset upload finish
        responses.add(responses.POST, url=self._url("resource/upload_finish"),
                      json={
                          'msg': "上传已完成，稍后可查看上传结果"
                      },
                      status=200)
        # Stub algo query
        responses.add(responses.GET, url=self._url("algorithm/query?id=ALGO001"),
                      match_querystring=True, json=[{
                          'id': "ALGO001",
                          'name': "TestAlgo1",
                          'description': "test",
                          'state': ResourceState.READY,
                          'visibility': ResourceVisibility.PUBLIC,
                          'upload_time': "2020-01-01 00:00",
                          'filename': "buzhongyao",
                          'is_zipfile': 1,
                          'file': "USER001/algos/ALGO001/buzhongyao",
                          'size': 250.41,
                          'creator_id': "USER001",
                          'node_id': "NODE001",
                          'owner': ["USER001"],
                          'tags': "tag1, tag2",
                          'train_params': '[{"test": "test"}]',
                          'evaluate_params': '[{"test": "test"}]',
                          'mirror_id': "MIRR002",
                          'follows_anylearn_norm': False,
                          'entrypoint_training': "python train.py",
                          'output_training': "output",
                          'entrypoint_evaluation': "python eval.py",
                          'output_evaluation': "output",
                      }],
                      status=200)
        # Stub model query
        responses.add(responses.GET, url=self._url("model/query?id=MODE001"),
                      match_querystring=True, json=[{
                          'id': "MODE001",
                          'name': "TestModel1",
                          'description': "test",
                          'state': ResourceState.READY,
                          'visibility': ResourceVisibility.PUBLIC,
                          'upload_time': "2020-01-01 00:00",
                          'filename': "buzhongyao",
                          'is_zipfile': 1,
                          'file': "USER001/models/MODE001/buzhongyao",
                          'size': 250.41,
                          'creator_id': "USER001",
                          'node_id': "NODE001",
                          'owner': ["USER001"],
                          'algorithm_id': "ALGO001",
                      }],
                      status=200)
        # Stub dset query
        responses.add(responses.GET, url=self._url("dataset/query?id=DSET001"),
                      match_querystring=True, json=[{
                          'id': "DSET001",
                          'name': "TestDset1",
                          'description': "test",
                          'state': ResourceState.READY,
                          'visibility': ResourceVisibility.PUBLIC,
                          'upload_time': "2020-01-01 00:00",
                          'filename': "buzhongyao",
                          'is_zipfile': 1,
                          'file': "USER001/datasets/DSET001/buzhongyao",
                          'size': 250.41,
                          'creator_id': "USER001",
                          'node_id': "NODE001",
                          'owner': ["USER001"],
                      }],
                      status=200)
        # Stub eval task creation
        responses.add(responses.POST, url=self._url("evaluate_task/add"),
                      json={'data': "EVAL001", 'message': "服务添加成功"},
                      status=200)
        # Stub eval task query
        responses.add(responses.GET, url=self._url("evaluate_task/query?id=EVAL001"),
                      match_querystring=True, json={
                          'id': "EVAL001",
                          'name': "TestEvaluateTask1",
                          'description': "test",
                          'state': EvaluateTaskState.RUNNING,
                          'visibility': 1,
                          'creator_id': "USER001",
                          'owner': ["USER001"],
                          'create_time': "2021-04-01 23:59:59",
                          'finish_time': "",
                          'params_list': '[{"model_id": "MODE001", "files": "DSET001", "evaluate_params": {"dataset": "$DSET001", "model_path": "$MODE001"}}]',
                          'secret_key': "SECRET",
                          'results': [
                              {
                                  'args': '{"dataset": "$DSET001", "model_path": "$MODE001"}',
                                  'envs': "",
                                  'files': "DSET001",
                                  'gpu_num': 1,
                                  'id': "EVRE001",
                                  'model_id': "MODE001",
                                  'results': "",
                              }
                          ],
                      },
                      status=200)

        eval_task, algo, model, dset = quick_evaluate(
            model_dir=FAKE_MODEL,
            algorithm_dir=FAKE_ALGO_OK,
            dataset_dir=FAKE_DSET,
            entrypoint="python main.py",
            output="output",
            resource_uploader=uploader,
            resource_polling=0.1,
            model_hyperparam_name="model_path",
            dataset_hyperparam_name="dataset",
            hyperparams={'test_param': "test"}
        )

        self.assertIsInstance(eval_task, EvaluateTask)
        self.assertEqual(eval_task.id, "EVAL001")
        self.assertTrue((FAKE_ALGO_OK / "anylearn_tools.py").exists())
        self.assertIsInstance(algo, Algorithm)
        self.assertEqual(algo.id, "ALGO001")
        self.assertIsInstance(model, Model)
        self.assertEqual(model.id, "MODE001")
        self.assertIsInstance(dset, Dataset)
        self.assertEqual(dset.id, "DSET001")

    @pytest.mark.skip()
    def test_quick_evaluate_ko_algo_missing_requirements_txt(self):
        with self.assertRaises(AnyLearnException) as ctx:
            eval_task = quick_evaluate(model_dir=FAKE_MODEL,
                                       algorithm_dir=FAKE_ALGO_KO,
                                       dataset_dir=FAKE_DSET,
                                       entrypoint="python main.py",
                                       output="output")
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnException)
        self.assertEqual(e.msg, "Missing 'requirements.txt' in algorithm folder")

    @pytest.mark.skip()
    @responses.activate
    def test_quick_evaluate_ko_mirror_not_found(self):
        responses.add(responses.GET, url=self._url("mirror/list"),
                      json=[
                          {
                              'id': "MIRR001",
                              'name': "Test001",
                              'requirements': "test==0.0.1",
                          },
                          {
                              'id': "MIRR002",
                              'name': "Test002",
                              'requirements': "test==0.0.2",
                          },
                      ],
                      status=200)
        with self.assertRaises(AnyLearnNotSupportedException) as ctx:
            eval_task = quick_evaluate(model_dir=FAKE_MODEL,
                                       algorithm_dir=FAKE_ALGO_OK,
                                       dataset_dir=FAKE_DSET,
                                       entrypoint="python main.py",
                                       output="output")
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnNotSupportedException)
        self.assertEqual(e.msg, "Container for `QUICKSTART` is not supported by the connected backend.")

    @pytest.mark.skip()
    @patch('anylearn.interfaces.resource.resource_uploader.ResourceUploader')
    @responses.activate
    def test_quick_evaluate_ko_algo_on_error(self, MockResourceUploader):
        @create_autospec
        def mock_upload_ok(resource_id, chunks):
            return True

        # Mock uploader with stubbed upload function
        uploader = MockResourceUploader()
        uploader.run = mock_upload_ok

        # Stub mirror list
        responses.add(responses.GET, url=self._url("mirror/list"),
                      json=[
                          {
                              'id': "MIRR001",
                              'name': "Test001",
                              'requirements': "test==0.0.1",
                          },
                          {
                              'id': "MIRR002",
                              'name': "QUICKSTART",
                              'requirements': "test==0.0.2",
                          },
                      ],
                      status=200)
        # Stub algo creation
        responses.add(responses.POST, url=self._url("algorithm/add"),
                      json={'data': "ALGO250", 'message': "算法添加成功"},
                      status=200)
        # Stub model creation
        responses.add(responses.POST, url=self._url("model/add"),
                      json={'data': "MODE001", 'message': "模型添加成功"},
                      status=200)
        # Stub dset creation
        responses.add(responses.POST, url=self._url("dataset/add"),
                      json={'data': "DSET001", 'message': "数据集添加成功"},
                      status=200)
        # Stub algo upload finish
        responses.add(responses.POST, url=self._url("resource/upload_finish"),
                      json={
                          'msg': "上传已完成，稍后可查看上传结果"
                      },
                      status=200)
        # Stub model upload finish
        responses.add(responses.POST, url=self._url("resource/upload_finish"),
                      json={
                          'msg': "上传已完成，稍后可查看上传结果"
                      },
                      status=200)
        # Stub dset upload finish
        responses.add(responses.POST, url=self._url("resource/upload_finish"),
                      json={
                          'msg': "上传已完成，稍后可查看上传结果"
                      },
                      status=200)
        # Stub algo query
        responses.add(responses.GET, url=self._url("algorithm/query?id=ALGO250"),
                      match_querystring=True, json=[{
                          'id': "ALGO250",
                          'name': "TestAlgo1",
                          'description': "test",
                          'state': ResourceState.ERROR,
                          'visibility': ResourceVisibility.PUBLIC,
                          'upload_time': "2020-01-01 00:00",
                          'filename': "buzhongyao",
                          'is_zipfile': 1,
                          'file': "USER001/algos/ALGO250/buzhongyao",
                          'size': 250.41,
                          'creator_id': "USER001",
                          'node_id': "NODE001",
                          'owner': ["USER001"],
                          'tags': "tag1, tag2",
                          'train_params': '[{"test": "test"}]',
                          'evaluate_params': '[{"test": "test"}]',
                          'mirror_id': "MIRR002",
                          'follows_anylearn_norm': False,
                          'entrypoint_training': "python train.py",
                          'output_training': "output",
                          'entrypoint_evaluation': "python eval.py",
                          'output_evaluation': "output",
                      }],
                      status=200)

        with self.assertRaises(AnyLearnException) as ctx:
            eval_task = quick_evaluate(model_dir=FAKE_MODEL,
                                       algorithm_dir=FAKE_ALGO_OK,
                                       dataset_dir=FAKE_DSET,
                                       entrypoint="python main.py",
                                       output="output",
                                       resource_uploader=uploader,
                                       resource_polling=0.1)
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnException)
        self.assertEqual(e.msg, "Error occured when uploading algorithm")

    @pytest.mark.skip()
    @patch('anylearn.interfaces.resource.resource_uploader.ResourceUploader')
    @responses.activate
    def test_quick_evaluate_ko_dset_on_error(self, MockResourceUploader):
        @create_autospec
        def mock_upload_ok(resource_id, chunks):
            return True

        # Mock uploader with stubbed upload function
        uploader = MockResourceUploader()
        uploader.run = mock_upload_ok

        # Stub mirror list
        responses.add(responses.GET, url=self._url("mirror/list"),
                      json=[
                          {
                              'id': "MIRR001",
                              'name': "Test001",
                              'requirements': "test==0.0.1",
                          },
                          {
                              'id': "MIRR002",
                              'name': "QUICKSTART",
                              'requirements': "test==0.0.2",
                          },
                      ],
                      status=200)
        # Stub algo creation
        responses.add(responses.POST, url=self._url("algorithm/add"),
                      json={'data': "ALGO001", 'message': "算法添加成功"},
                      status=200)
        # Stub model creation
        responses.add(responses.POST, url=self._url("model/add"),
                      json={'data': "MODE001", 'message': "模型添加成功"},
                      status=200)
        # Stub dset creation
        responses.add(responses.POST, url=self._url("dataset/add"),
                      json={'data': "DSET250", 'message': "数据集添加成功"},
                      status=200)
        # Stub algo upload finish
        responses.add(responses.POST, url=self._url("resource/upload_finish"),
                      json={
                          'msg': "上传已完成，稍后可查看上传结果"
                      },
                      status=200)
        # Stub model upload finish
        responses.add(responses.POST, url=self._url("resource/upload_finish"),
                      json={
                          'msg': "上传已完成，稍后可查看上传结果"
                      },
                      status=200)
        # Stub dset upload finish
        responses.add(responses.POST, url=self._url("resource/upload_finish"),
                      json={
                          'msg': "上传已完成，稍后可查看上传结果"
                      },
                      status=200)
        # Stub model query
        responses.add(responses.GET, url=self._url("model/query?id=MODE001"),
                      match_querystring=True, json=[{
                          'id': "MODE001",
                          'name': "TestModel1",
                          'description': "test",
                          'state': ResourceState.READY,
                          'visibility': ResourceVisibility.PUBLIC,
                          'upload_time': "2020-01-01 00:00",
                          'filename': "buzhongyao",
                          'is_zipfile': 1,
                          'file': "USER001/models/MODE001/buzhongyao",
                          'size': 250.41,
                          'creator_id': "USER001",
                          'node_id': "NODE001",
                          'owner': ["USER001"],
                          'algorithm_id': "ALGO001",
                      }],
                      status=200)
        # Stub algo query
        responses.add(responses.GET, url=self._url("algorithm/query?id=ALGO001"),
                      match_querystring=True, json=[{
                          'id': "ALGO001",
                          'name': "TestAlgo1",
                          'description': "test",
                          'state': ResourceState.READY,
                          'visibility': ResourceVisibility.PUBLIC,
                          'upload_time': "2020-01-01 00:00",
                          'filename': "buzhongyao",
                          'is_zipfile': 1,
                          'file': "USER001/algos/ALGO001/buzhongyao",
                          'size': 250.41,
                          'creator_id': "USER001",
                          'node_id': "NODE001",
                          'owner': ["USER001"],
                          'tags': "tag1, tag2",
                          'train_params': '[{"test": "test"}]',
                          'evaluate_params': '[{"test": "test"}]',
                          'mirror_id': "MIRR002",
                          'follows_anylearn_norm': False,
                          'entrypoint_training': "python train.py",
                          'output_training': "output",
                          'entrypoint_evaluation': "python eval.py",
                          'output_evaluation': "output",
                      }],
                      status=200)
        # Stub dset query
        responses.add(responses.GET, url=self._url("dataset/query?id=DSET250"),
                      match_querystring=True, json=[{
                          'id': "DSET250",
                          'name': "TestDset1",
                          'description': "test",
                          'state': ResourceState.ERROR,
                          'visibility': ResourceVisibility.PUBLIC,
                          'upload_time': "2020-01-01 00:00",
                          'filename': "buzhongyao",
                          'is_zipfile': 1,
                          'file': "USER001/datasets/DSET250/buzhongyao",
                          'size': 250.41,
                          'creator_id': "USER001",
                          'node_id': "NODE001",
                          'owner': ["USER001"],
                      }],
                      status=200)

        with self.assertRaises(AnyLearnException) as ctx:
            eval_task = quick_evaluate(model_dir=FAKE_MODEL,
                                       algorithm_dir=FAKE_ALGO_OK,
                                       dataset_dir=FAKE_DSET,
                                       entrypoint="python main.py",
                                       output="output",
                                       resource_uploader=uploader,
                                       resource_polling=0.1)
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnException)
        self.assertEqual(e.msg, "Error occured when uploading dataset")

    @pytest.mark.skip()
    @patch('anylearn.interfaces.resource.resource_uploader.ResourceUploader')
    @responses.activate
    def test_quick_evaluate_ko_model_on_error(self, MockResourceUploader):
        @create_autospec
        def mock_upload_ok(resource_id, chunks):
            return True

        # Mock uploader with stubbed upload function
        uploader = MockResourceUploader()
        uploader.run = mock_upload_ok

        # Stub mirror list
        responses.add(responses.GET, url=self._url("mirror/list"),
                      json=[
                          {
                              'id': "MIRR001",
                              'name': "Test001",
                              'requirements': "test==0.0.1",
                          },
                          {
                              'id': "MIRR002",
                              'name': "QUICKSTART",
                              'requirements': "test==0.0.2",
                          },
                      ],
                      status=200)
        # Stub algo creation
        responses.add(responses.POST, url=self._url("algorithm/add"),
                      json={'data': "ALGO001", 'message': "算法添加成功"},
                      status=200)
        # Stub model creation
        responses.add(responses.POST, url=self._url("model/add"),
                      json={'data': "MODE250", 'message': "模型添加成功"},
                      status=200)
        # Stub dset creation
        responses.add(responses.POST, url=self._url("dataset/add"),
                      json={'data': "DSET001", 'message': "数据集添加成功"},
                      status=200)
        # Stub algo upload finish
        responses.add(responses.POST, url=self._url("resource/upload_finish"),
                      json={
                          'msg': "上传已完成，稍后可查看上传结果"
                      },
                      status=200)
        # Stub model upload finish
        responses.add(responses.POST, url=self._url("resource/upload_finish"),
                      json={
                          'msg': "上传已完成，稍后可查看上传结果"
                      },
                      status=200)
        # Stub dset upload finish
        responses.add(responses.POST, url=self._url("resource/upload_finish"),
                      json={
                          'msg': "上传已完成，稍后可查看上传结果"
                      },
                      status=200)
        # Stub algo query
        responses.add(responses.GET, url=self._url("algorithm/query?id=ALGO001"),
                      match_querystring=True, json=[{
                          'id': "ALGO001",
                          'name': "TestAlgo1",
                          'description': "test",
                          'state': ResourceState.READY,
                          'visibility': ResourceVisibility.PUBLIC,
                          'upload_time': "2020-01-01 00:00",
                          'filename': "buzhongyao",
                          'is_zipfile': 1,
                          'file': "USER001/algos/ALGO001/buzhongyao",
                          'size': 250.41,
                          'creator_id': "USER001",
                          'node_id': "NODE001",
                          'owner': ["USER001"],
                          'tags': "tag1, tag2",
                          'train_params': '[{"test": "test"}]',
                          'evaluate_params': '[{"test": "test"}]',
                          'mirror_id': "MIRR002",
                          'follows_anylearn_norm': False,
                          'entrypoint_training': "python train.py",
                          'output_training': "output",
                          'entrypoint_evaluation': "python eval.py",
                          'output_evaluation': "output",
                      }],
                      status=200)
        # Stub model query
        responses.add(responses.GET, url=self._url("model/query?id=MODE250"),
                      match_querystring=True, json=[{
                          'id': "MODE250",
                          'name': "TestModel1",
                          'description': "test",
                          'state': ResourceState.ERROR,
                          'visibility': ResourceVisibility.PUBLIC,
                          'upload_time': "2020-01-01 00:00",
                          'filename': "buzhongyao",
                          'is_zipfile': 1,
                          'file': "USER001/models/MODE250/buzhongyao",
                          'size': 250.41,
                          'creator_id': "USER001",
                          'node_id': "NODE001",
                          'owner': ["USER001"],
                          'algorithm_id': "ALGO001",
                      }],
                      status=200)

        with self.assertRaises(AnyLearnException) as ctx:
            eval_task = quick_evaluate(model_dir=FAKE_MODEL,
                                       algorithm_dir=FAKE_ALGO_OK,
                                       dataset_dir=FAKE_DSET,
                                       entrypoint="python main.py",
                                       output="output",
                                       resource_uploader=uploader,
                                       resource_polling=0.1)
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnException)
        self.assertEqual(e.msg, "Error occured when uploading model")



import hashlib
from pathlib import Path
import random
import shutil
import string
from typing import Optional, Union

from ..config import AnylearnConfig
from ..utils import logger
from ..utils.errors import AnyLearnMissingParamException


def make_name_by_path(path: Union[str, Path]) -> str:
    if not path:
        raise AnyLearnMissingParamException("`path` required.")
    path = Path(path)
    basename = path.name
    suffix = hashlib.sha1(str(path).encode('utf-8')).hexdigest()[:8]
    return f"{basename}-{suffix}"


def generate_random_name() -> str:
    return ''.join(random.sample(string.ascii_lowercase + string.digits, 8))


def _check_resource_input(id: Optional[str]=None,
                          dir_path: Optional[Union[str, Path]]=None,
                          archive_path: Optional[str]=None):
    if not any([id, dir_path, archive_path]):
        raise AnyLearnMissingParamException((
            "At least one of the parameters "
            "['id', 'dir_path', 'archive_path'] "
            "should be specified."
        ))


def _get_or_create_resource_archive(name,
                                    dir_path: Optional[Union[str, Path]]=None,
                                    archive_path: Optional[str]=None):
    logger.info("Packaging resources...")
    if not archive_path or not Path(archive_path).exists():
        archive_path = shutil.make_archive(
            AnylearnConfig.workspace_path / name,
            "zip",
            dir_path
        )
    return archive_path


def _get_archive_checksum(archive_path: str, buffer_size: int=65536):
    checksum = hashlib.blake2b()
    with open(archive_path, "rb") as f:
        while True:
            chunk = f.read(buffer_size)
            if not chunk:
                break
            checksum.update(chunk)
    return checksum.hexdigest()



from __future__ import annotations
from datetime import datetime
from typing import Optional

'''
所有与AnyLearn Project相关的接口
'''
from anylearn.utils.api import url_base, get_with_token, post_with_token, put_with_token, delete_with_token
from anylearn.utils.errors import AnyLearnException, AnyLearnMissingParamException
from anylearn.utils import no_none_filter
from anylearn.interfaces.base import BaseObject
from anylearn.interfaces.train_task import TrainTask


class ProjectVisibility:
    """
    项目可见性标识：

    - -1(HIDDEN)表示不可见
    - 1(PRIVATE)表示仅创建者可见
    - 2(PROTECTED)表示所有者可见
    - 3(PUBLIC)表示公开
    """
    HIDDEN = -1
    PRIVATE = 1
    PROTECTED = 2
    PUBLIC = 3


class Project(BaseObject):
    """
    AnyLearn项目类，以方法映射项目CRUD相关接口

    Attributes
    ----------
    id
        项目的唯一标识符，自动生成，由PROJ+uuid1生成的编码中后28个有效位（小写字母和数字）组成
    name
        项目名称（非空 长度1~50）
    description
        项目描述（可为空 长度最大255）
    visibility
        项目的可见性（默认值1）
    create_time
        创建时间
    update_time
        更新时间
    creator_id
        创建者的ID
    datasets
        项目实用的数据集,以逗号分隔数据集的ID拼成的字符串，无多余空格
    owner
        项目的所有者，以逗号分隔用户的ID拼成的字符串，无多余空格
    load_detail
        初始化时是否加载详情
    """

    _fields = {
        # 资源创建/更新请求包体中必须包含且不能为空的字段
        'required': {
            'create': ['name'],
            'update': ['id', 'name'],
        },
        # 资源创建/更新请求包体中包含的所有字段
        'payload': {
            'create': ['id', 'name', 'description', 'datasets', 'visibility',
                       'owner'],
            'update': ['id', 'name', 'description', 'datasets', 'visibility',
                       'owner'],
        },
    }
    """
    创建/更新对象时：

    - 必须包含且不能为空的字段 :obj:`_fields['required']`
    - 所有字段 :obj:`_fields['payload']`
    """

    def __init__(self,
                 id: Optional[str]=None,
                 name: Optional[str]=None,
                 description: Optional[str]=None,
                 visibility=3,
                 create_time: Optional[datetime]=None,
                 update_time: Optional[datetime]=None,
                 creator_id: Optional[str]=None,
                 datasets: Optional[list]=None,
                 owner: Optional[list]=None,
                 load_detail=False):
        """
        Parameters
        ----------
        id
            项目的唯一标识符，自动生成，由PROJ+uuid1生成的编码中后28个有效位（小写字母和数字）组成
        name
            项目名称（非空 长度1~50）
        description
            项目描述（可为空 长度最大255）
        visibility
            项目的可见性（默认值1）
        create_time
            创建时间
        update_time
            更新时间
        creator_id
            创建者的ID
        datasets
            项目实用的数据集,以逗号分隔数据集的ID拼成的字符串，无多余空格
        owner
            项目的所有者，以逗号分隔用户的ID拼成的字符串，无多余空格
        load_detail
            初始化时是否加载详情
        """
        self.id = id
        self.name = name
        self.description = description
        self.visibility = visibility
        self.create_time = create_time
        self.update_time = update_time
        self.creator_id = creator_id
        self.datasets = datasets
        self.owner = owner
        super().__init__(id=id, load_detail=load_detail)

    def __eq__(self, other: Project) -> bool:
        if not isinstance(other, Project):
            return NotImplemented
        return all([
            self.id == other.id,
            self.name == other.name,
            self.description == other.description,
            # self.visibility == other.visibility,
            # self.create_time == other.create_time,
            # self.update_time == other.update_time,
            # self.creator_id == other.creator_id,
            # self.datasets == other.datasets,
            # self.owner == other.owner,
        ])

    @classmethod
    def get_list(cls):
        """
        获取训练项目列表
        
        Returns
        -------
        List [Project]
            训练项目对象的集合。
        """
        res = get_with_token(f"{url_base()}/project/list")
        if res is None or not isinstance(res, list):
            raise AnyLearnException("请求未能得到有效响应")
        return [
            Project(id=item['id'], name=item['name'],
                    description=item['description'],
                    create_time=item['create_time'],
                    update_time=item['update_time'])
            for item in res
        ]

    def get_detail(self):
        """
        获取训练项目详细信息

        - 对象属性 :obj:`id` 应为非空

        Returns
        -------
        Project
            训练项目对象。
        """
        self._check_fields(required=['id'])
        res = get_with_token(f"{url_base()}/project/query",
                             params={'id': self.id})
        if not res or not isinstance(res, list):
            raise AnyLearnException("请求未能得到有效响应")
        res = res[0]
        self.__init__(id=res['id'], name=res['name'],
                      description=res['description'],
                      visibility=res['visibility'],
                      create_time=res['create_time'],
                      update_time=res['update_time'],
                      creator_id=res['creator_id'],
                      datasets=res['datasets'],
                      owner=res['owner'])

    @classmethod
    def get_my_default_project(cls) -> Project:
        res = get_with_token(f"{url_base()}/project/default")
        if not isinstance(res, list):
            raise AnyLearnException("请求未能得到有效响应")
        res = res[0]
        return Project(id=res['id'], name=res['name'],
                       description=res['description'],
                       visibility=res['visibility'],
                       create_time=res['create_time'],
                       update_time=res['update_time'],
                       creator_id=res['creator_id'],
                       datasets=res['datasets'],
                       owner=res['owner'])

    @classmethod
    def create_my_default_project(cls) -> Project:
        res = get_with_token(f"{url_base()}/project/default")
        if not isinstance(res, dict) or 'data' not in res:
            raise AnyLearnException("请求未能得到有效响应")
        project_id = res['data']
        return Project(id=project_id, load_detail=True)

    def get_train_tasks(self, load_detail=False):
        """
        获取训练项目的训练任务列表
        
        - 对象属性 :obj:`id` 应为非空

        Returns
        -------
        List [TrainTask]
            训练任务的集合。
        """
        self._check_fields(required=['id'])
        res = get_with_token(f"{url_base()}/train_task/list",
                             params={'project_id': self.id})
        if res is None or not isinstance(res, list):
            raise AnyLearnException("请求未能得到有效响应")
        return [
            TrainTask(id=item['id'], name=item['name'],
                      description=item['description'], state=item['state'],
                      secret_key=item['secret_key'],
                      results_id=item['results_id'],
                      create_time=item['create_time'],
                      finish_time=item['finish_time'],
                      load_detail=load_detail)
            for item in res
        ]

    def _namespace(self):
        return "project"



"""alter_train_profile_datasetid_nullable

Revision ID: ed56a42729b7
Revises: c1ee06d84c12
Create Date: 2021-07-26 17:21:01.231388

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'ed56a42729b7'
down_revision = 'c1ee06d84c12'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('local_train_profile', schema=None) as batch_op:
        batch_op.alter_column('dataset_id',
               existing_type=sa.VARCHAR(length=32),
               nullable=True)

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('local_train_profile', schema=None) as batch_op:
        batch_op.alter_column('dataset_id',
               existing_type=sa.VARCHAR(length=32),
               nullable=False)

    # ### end Alembic commands ###



from abc import ABC, abstractmethod
import asyncio

import aiohttp
import requests
from tqdm import tqdm

from anylearn.config import AnylearnConfig
from anylearn.utils import logger
from anylearn.utils.api import url_base


class ResourceUploader(ABC):
    """
    资源上传工具接口
    """

    @abstractmethod
    def run(self, resource_id: str, chunks: list):
        """
        执行资源上传,自定义资源上传需实现此方法

        Parameters
        ----------
        resource_id
            资源ID
        chunks
            被切割后的文件内容列表
        """
        raise NotImplementedError


class AsyncResourceUploader(ResourceUploader):
    """
    资源异步上传工具类
    """

    def __init__(self, on_start=None, on_progress=None):
        self.__counter = 0
        self.__total = 0
        self.__on_start = on_start
        self.__on_progress = on_progress
        super().__init__()

    def run(self, resource_id: str, chunks: list):
        """
        执行资源上传
        """
        return asyncio.run(self.__run(resource_id=resource_id,
                                      chunks=chunks))

    async def __run(self, resource_id: str, chunks: list):
        """执行异步上传请求

        Args:
            resource_id: 后端资源ID
            chunks: 被切割后的文件内容列表

        Returns:
            bool: 上传成功与否
        
        Raises:
            ClientResponseError: 当上传过程中出错时由aiohttp包（默认情况）抛出错误
        """

        self.__counter = 0
        self.__total = len(chunks)
        if callable(self.__on_start):
            self.__on_start(self.__total)

        headers = {'Authorization': f"Bearer {AnylearnConfig.token}"}
        async with aiohttp.ClientSession(headers=headers) as session:
            tasks = [self.__do_upload(session=session,
                                      resource_id=resource_id,
                                      chunk=chunk,
                                      chunk_index=i)
                        for (i, chunk) in enumerate(chunks)]
            res = await asyncio.gather(*tasks)
        return all(res)

    async def __do_upload(self,
                          session: aiohttp.ClientSession,
                          resource_id: str,
                          chunk: str,
                          chunk_index: int):
        res = await session.request(method="POST",
                                    url=f"{url_base()}/resource/upload",
                                    raise_for_status=True,
                                    data={
                                        'file_id': resource_id,
                                        'file': chunk,
                                        'chunk': str(chunk_index),
                                    })
        self.__counter += 1
        if callable(self.__on_progress):
            self.__on_progress(self.__counter, self.__total)
        return res.ok


class SyncResourceUploader(ResourceUploader):
    def run(self, resource_id: str, chunks: list):
        for i, chunk in enumerate(tqdm(chunks)):
            url = f"{url_base()}/resource/upload"
            headers = {'Authorization': f"Bearer {AnylearnConfig.token}"}
            files = {'file': chunk}
            data = {'file_id': resource_id, 'chunk': str(i)}
            res = requests.post(url, headers=headers, files=files, data=data)
            res.raise_for_status()



from typing import Optional

from anylearn.utils.api import url_base, get_with_token
from anylearn.utils.errors import AnyLearnException

class Mirror:
    """
    AnyLearn镜像类，仅提供镜像列表接口

    Attributes
    ----------
    id
        镜像的唯一标识符，自动生成，由MIRR+uuid1生成的编码中后28个有效位（小写字母和数字）组成
    name
        镜像的名称
    requirements
        依赖包
    """

    def __init__(self,
                 id: Optional[str]=None,
                 name: Optional[str]=None,
                 requirements: Optional[str]=None):
        """
        AnyLearn镜像类，仅提供镜像列表接口

        Parameters
        ----------
        id
            镜像的唯一标识符，自动生成，由MIRR+uuid1生成的编码中后28个有效位（小写字母和数字）组成
        name
            镜像的名称
        requirements
            依赖包
        """
        self.id = id
        self.name = name
        self.requirements = requirements

    @classmethod
    def get_list(cls) -> list:
        """
        获取镜像列表
        
        Returns
        -------
        List [Mirror]
            镜像对象的集合。
        """
        res = get_with_token(f"{url_base()}/mirror/list")
        if res is None or not isinstance(res, list):
            raise AnyLearnException("请求未能得到有效响应")
        return [
            Mirror(id=m['id'], name=m['name'], requirements=m['requirements'])
            for m in res
        ]

    def __str__(self):
        return f"Mirror(name={self.name})"

    def __repr__(self):
        d = self.__dict__
        kv = ", ".join([f"{k}={d[k]!r}" for k in d])
        return f"{self.__class__.__name__}({kv})"
